@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
-"""Automated Tool for Optimized Modelling (ATOM).
-
+"""
+Automated Tool for Optimized Modelling (ATOM)
 Author: Mavs
 Description: Module containing the parent class for the trainers.
 
@@ -17,14 +17,13 @@ from skopt.callbacks import DeadlineStopper, DeltaXStopper, DeltaYStopper
 
 # Own modules
 from .branch import Branch
-from .models import MODEL_LIST, CustomModel
+from .models import MODELS, CustomModel
 from .basepredictor import BasePredictor
 from .data_cleaning import BaseTransformer
-from .basemodel import BaseModel
 from .utils import (
-    SEQUENCE, OPTIONAL_PACKAGES, ONLY_CLASS, ONLY_REG, lst,
-    dct, time_to_str, get_acronym, get_scorer, get_best_score,
-    check_scaling, delete, PlotCallback, CustomDict,
+    SEQUENCE, OPTIONAL_PACKAGES, lst, time_to_str, is_multidim,
+    is_sparse, get_custom_scorer, get_best_score, check_scaling,
+    delete, PlotCallback, CustomDict,
 )
 
 
@@ -85,9 +84,9 @@ class BaseTrainer(BaseTransformer, BasePredictor):
     est_params: dict or None, optional (default=None)
         Additional parameters for the estimators. See the corresponding
         documentation for the available options. For multiple models,
-        use the acronyms as key and a dict of the parameters as value.
-        Add _fit to the parameter's name to pass it to the fit method
-        instead of the initializer.
+        use the acronyms as key (or 'all' for all models) and a dict
+        of the parameters as value. Add _fit to the parameter's name
+        to pass it to the fit method instead of the initializer.
 
     bo_params: dict or None, optional (default=None)
         Additional parameters to for the BO. These can include:
@@ -114,13 +113,13 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                 Number of folds for the cross-validation. If 1, the
                 training set is randomly split in a (sub)train and
                 validation set.
-            - callbacks: callable or sequence, optional (default=None)
+            - callback: callable or list of callables, optional (default=None)
                 Callbacks for the BO.
-            - dimensions: dict, sequence or None, optional (default=None)
-                Custom hyperparameter space for the BO. Can be an
-                array to share the same dimensions across models
-                or a dictionary with the model names as key. If
-                None, ATOM's predefined dimensions are used.
+            - dimensions: dict, list or None, optional (default=None)
+                Custom hyperparameter space for the BO. Can be a list
+                to share the same dimensions across models or a dict
+                with the model names as key (or `all` for all models).
+                If None, ATOM's predefined dimensions are used.
             - plot: bool, optional (default=False)
                 Whether to plot the BO's progress as it runs.
                 Creates a canvas with two plots: the first plot
@@ -169,7 +168,7 @@ class BaseTrainer(BaseTransformer, BasePredictor):
 
     random_state: int or None, optional (default=None)
         Seed used by the random number generator. If None, the random
-        number generator is the `RandomState` used by `numpy.random`.
+        number generator is the `RandomState` used by `np.random`.
 
     """
 
@@ -188,56 +187,83 @@ class BaseTrainer(BaseTransformer, BasePredictor):
         )
 
         # Parameter attributes
-        self._models = CustomDict({model: model for model in lst(models)})
-        if not isinstance(metric, CustomDict):
-            self._metric = CustomDict({metric: metric for metric in lst(metric)})
-        else:
-            self._metric = metric
+        self._models = models
+        self._metric = metric
         self.greater_is_better = greater_is_better
         self.needs_proba = needs_proba
         self.needs_threshold = needs_threshold
         self.n_calls = n_calls
         self.n_initial_points = n_initial_points
-        self.est_params = dct(est_params)
-        self.bo_params = dct(bo_params)
+        self.est_params = est_params
+        self.bo_params = bo_params
         self.n_bootstrap = n_bootstrap
 
         # Branching attributes
-        self._current = "master"  # Current (and only) branch
-        self._branches = {self._current: Branch(self, self._current)}
+        self.index = True
+        self.holdout = None
+        self._current = "master"
+        self._branches = CustomDict({self._current: Branch(self, self._current)})
 
         # Training attributes
         self.task = None
         self.scaled = None
-        self.errors = {}
+        self._bo = {"base_estimator": "GP", "cv": 1, "callback": [], "kwargs": {}}
+        self._errors = CustomDict()
 
-        # BO attributes
-        self._base_estimator = None
-        self._cv = 5
-        self._callbacks = []
-        self._bo_kwargs = {}
+    @staticmethod
+    def _prepare_metric(metric, **kwargs):
+        """Check the validity of the metric."""
+        metric_params = {}
+        for key, value in kwargs.items():
+            if isinstance(value, SEQUENCE):
+                if len(value) != len(metric):
+                    raise ValueError(
+                        f"Invalid value for the {key} parameter. Its length "
+                        "should be equal to the number of metrics, got "
+                        f"len(metric)={len(metric)} and len({key})={len(value)}."
+                    )
+            else:
+                metric_params[key] = [value for _ in metric]
 
-    def _check_parameters(self):
+        metric_dict = CustomDict()
+        for args in zip(metric, *metric_params.values()):
+            scorer = get_custom_scorer(*args)
+            metric_dict[scorer.name] = scorer
+
+        return metric_dict
+
+    def _prepare_parameters(self):
         """Check the validity of the input parameters."""
-        if self.mapping is None:
-            self.mapping = {str(v): v for v in sorted(self.y.unique())}
-
-        if self.scaled is None:
+        if self.scaled is None and not is_multidim(self.X) and not is_sparse(self.X):
             self.scaled = check_scaling(self.X)
 
         # Create model subclasses ================================== >>
 
         # If left to default, select all predefined models per task
-        if None in self._models:
-            if self.goal.startswith("class"):
-                models = [m(self) for m in MODEL_LIST if m.acronym not in ONLY_REG]
+        if self._models is None:
+            if self.goal == "class":
+                models = [m(self) for m in MODELS.values() if "class" in m.goal]
             else:
-                models = [m(self) for m in MODEL_LIST if m.acronym not in ONLY_CLASS]
+                models = [m(self) for m in MODELS.values() if "reg" in m.goal]
         else:
             models = []
-            for m in self._models:
+            for m in lst(self._models):
                 if isinstance(m, str):
-                    acronym = get_acronym(m, must_be_equal=False)
+                    if is_multidim(self.X):
+                        raise ValueError(
+                            "Multidimensional datasets are not supported by ATOM's "
+                            "predefined models. Refer to the documentation for the "
+                            "use of custom models."
+                        )
+
+                    # Get the acronym from the called model
+                    names = [n for n in MODELS if m.lower().startswith(n.lower())]
+                    if not names:
+                        raise ValueError(
+                            f"Unknown model: {m}. Choose from: {', '.join(MODELS)}."
+                        )
+                    else:
+                        acronym = names[0]
 
                     # Check if packages for non-sklearn models are available
                     if acronym in OPTIONAL_PACKAGES:
@@ -249,41 +275,46 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                                 "package. Make sure it is installed."
                             )
 
+                    models.append(MODELS[acronym](self, acronym + m[len(acronym):]))
+
                     # Check for regression/classification-only models
-                    if self.goal.startswith("class") and acronym in ONLY_REG:
+                    if self.goal == "class" and self.goal not in models[-1].goal:
                         raise ValueError(
                             f"The {acronym} model can't perform classification tasks!"
                         )
-                    elif self.goal.startswith("reg") and acronym in ONLY_CLASS:
+                    elif self.goal == "reg" and self.goal not in models[-1].goal:
                         raise ValueError(
                             f"The {acronym} model can't perform regression tasks!"
                         )
 
-                    models.append(MODEL_LIST[acronym](self, acronym + m[len(acronym):]))
-
-                elif not isinstance(m, BaseModel):  # Model is custom estimator
+                else:  # Model is a custom estimator
                     models.append(CustomModel(self, estimator=m))
 
-                else:  # Model is already a model subclass (can happen with reruns)
-                    models.append(m)
+        names = [m.name for m in models]
+        if len(set(names)) != len(names):
+            raise ValueError(
+                "Invalid value for the models parameter. It seems there are "
+                "duplicate models. Add a tag to a model's acronym to train two "
+                "different models with the same estimator, e.g. models=['LR1', 'LR2']."
+            )
 
-        self._models = CustomDict({m.name: m for m in models})
+        self._models = CustomDict({name: model for name, model in zip(names, models)})
 
         # Define scorer ============================================ >>
 
         # Assign default scorer
-        if None in self._metric:
+        if self._metric is None:
             if self.task.startswith("bin"):
-                self._metric = CustomDict(f1=get_scorer("f1"))
+                self._metric = CustomDict(f1=get_custom_scorer("f1"))
             elif self.task.startswith("multi"):
-                self._metric = CustomDict(f1_weighted=get_scorer("f1_weighted"))
+                self._metric = CustomDict(f1_weighted=get_custom_scorer("f1_weighted"))
             else:
-                self._metric = CustomDict(r2=get_scorer("r2"))
+                self._metric = CustomDict(r2=get_custom_scorer("r2"))
 
         # Ignore if it's the same scorer as previous call
-        elif not all([hasattr(m, "name") for m in self._metric]):
+        elif not isinstance(self._metric, CustomDict):
             self._metric = self._prepare_metric(
-                metric=self._metric,
+                metric=lst(self._metric),
                 greater_is_better=self.greater_is_better,
                 needs_proba=self.needs_proba,
                 needs_threshold=self.needs_threshold,
@@ -300,107 +331,25 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                     f"(models)={len(self._models)} and len({param})={len(p)}."
                 )
 
-            for i, model in enumerate(self._models):
-                if param in ("n_calls", "bootstrap") and p[i % len(p)] < 0:
+            for i, model in enumerate(self._models.values()):
+                value = p[i % len(p)]
+                if param == "n_calls" and (value == 1 or value < 0):
                     raise ValueError(
                         f"Invalid value for the {param} parameter. "
-                        f"Value should be >=0, got {p[i % len(p)]}."
+                        f"Value should be >=2, got {value}."
                     )
-                elif param == "n_initial_points" and p[i % len(p)] <= 0:
+                elif param == "n_initial_points" and value <= 0:
                     raise ValueError(
                         f"Invalid value for the {param} parameter. "
-                        f"Value should be >0, got {p[i % len(p)]}."
+                        f"Value should be >=1, got {value}."
+                    )
+                elif param == "n_bootstrap" and value < 0:
+                    raise ValueError(
+                        f"Invalid value for the {param} parameter. "
+                        f"Value should be >=0, got {value}."
                     )
 
-                setattr(model, "_" + param, p[i % len(p)])
-
-        # Prepare bo parameters ===================================== >>
-
-        # Choose a base estimator (GP is chosen as default)
-        self._base_estimator = self.bo_params.get("base_estimator", "GP")
-        if isinstance(self._base_estimator, str):
-            if self._base_estimator.lower() not in ("gp", "et", "rf", "gbrt"):
-                raise ValueError(
-                    f"Invalid value for the base_estimator parameter, got "
-                    f"{self._base_estimator}. Value should be one of: 'GP', "
-                    f"'ET', 'RF', 'GBRT'."
-                )
-
-        if self.bo_params.get("callbacks"):
-            self._callbacks = lst(self.bo_params["callbacks"])
-
-        if "max_time" in self.bo_params:
-            if self.bo_params["max_time"] <= 0:
-                raise ValueError(
-                    "Invalid value for the max_time parameter. "
-                    f"Value should be >0, got {self.bo_params['max_time']}."
-                )
-            self._callbacks.append(DeadlineStopper(self.bo_params["max_time"]))
-
-        if "delta_x" in self.bo_params:
-            if self.bo_params["delta_x"] < 0:
-                raise ValueError(
-                    "Invalid value for the delta_x parameter. "
-                    f"Value should be >=0, got {self.bo_params['delta_x']}."
-                )
-            self._callbacks.append(DeltaXStopper(self.bo_params["delta_x"]))
-
-        if "delta_y" in self.bo_params:
-            if self.bo_params["delta_y"] < 0:
-                raise ValueError(
-                    "Invalid value for the delta_y parameter. "
-                    f"Value should be >=0, got {self.bo_params['delta_y']}."
-                )
-            self._callbacks.append(DeltaYStopper(self.bo_params["delta_y"], n_best=5))
-
-        if self.bo_params.get("plot"):
-            self._callbacks.append(PlotCallback(self))
-
-        if "cv" in self.bo_params:
-            if self.bo_params["cv"] <= 0:
-                raise ValueError(
-                    "Invalid value for the max_time parameter. "
-                    f"Value should be >=0, got {self.bo_params['cv']}."
-                )
-            self._cv = self.bo_params["cv"]
-
-        if "early_stopping" in self.bo_params:
-            if self.bo_params["early_stopping"] <= 0:
-                raise ValueError(
-                    "Invalid value for the early_stopping parameter. "
-                    f"Value should be >=0, got {self.bo_params['early_stopping']}."
-                )
-            for model in self._models:
-                model._early_stopping = self.bo_params["early_stopping"]
-
-        # Add custom dimensions to every model subclass
-        if self.bo_params.get("dimensions"):
-            for name, model in self._models.items():
-                # If not dict, the dimensions are for all models
-                if not isinstance(self.bo_params["dimensions"], dict):
-                    model._dimensions = self.bo_params["dimensions"]
-                else:
-                    # Dimensions for every specific model
-                    for key, value in self.bo_params["dimensions"].items():
-                        # Parameters for this model only
-                        if key.lower() == name.lower():
-                            model._dimensions = value
-                            break
-
-        kwargs = [
-            "base_estimator",
-            "max_time",
-            "delta_x",
-            "delta_y",
-            "early_stopping",
-            "cv",
-            "callbacks",
-            "dimensions",
-            "plot",
-        ]
-
-        # The remaining bo_params are added as kwargs to the optimizer
-        self._bo_kwargs = {k: v for k, v in self.bo_params.items() if k not in kwargs}
+                setattr(model, "_" + param, value)
 
         # Prepare est_params ======================================= >>
 
@@ -409,7 +358,7 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                 params = {}
                 for key, value in self.est_params.items():
                     # Parameters for this model only
-                    if key.lower() == name.lower():
+                    if key.lower() == name.lower() or key.lower() == "all":
                         params.update(value)
                     # Parameters for all models
                     elif key not in self._models:
@@ -421,34 +370,102 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                     else:
                         model._est_params[key] = value
 
-    @staticmethod
-    def _prepare_metric(metric, **kwargs):
-        """Return a list of scorers given the parameters."""
-        metric_params = {}
-        for key, value in kwargs.items():
-            if isinstance(value, SEQUENCE):
-                if len(value) != len(metric):
+        # Prepare bo params ======================================== >>
+
+        base_estimators = ["GP", "RF", "ET", "GBRT"]
+        exc = ["max_time", "delta_x", "delta_y", "plot", "early_stopping", "dimensions"]
+
+        if self.bo_params:
+            if self.bo_params.get("base_estimator"):
+                self._bo["base_estimator"] = self.bo_params["base_estimator"]
+                if isinstance(self._bo["base_estimator"], str):
+                    if self._bo["base_estimator"].upper() not in base_estimators:
+                        raise ValueError(
+                            "Invalid value for the base_estimator parameter, "
+                            f"got {self.bo_params['base_estimator']}. Choose "
+                            f"from: {', '.join(base_estimators)}."
+                        )
+
+            if self.bo_params.get("callback"):
+                self._bo["callback"].extend(lst(self.bo_params["callback"]))
+
+            if "max_time" in self.bo_params:
+                if self.bo_params["max_time"] <= 0:
                     raise ValueError(
-                        "Invalid value for the greater_is_better parameter. Length "
-                        "should be equal to the number of metrics, got len(metric)="
-                        f"{len(metric)} and len({key})={len(value)}."
+                        "Invalid value for the max_time parameter. Value "
+                        f"should be >0, got {self.bo_params['max_time']}."
                     )
-            else:
-                metric_params[key] = [value for _ in metric]
+                max_time_callback = DeadlineStopper(self.bo_params["max_time"])
+                self._bo["callback"].append(max_time_callback)
 
-        metric_dict = CustomDict()
-        for args in zip(metric, *metric_params.values()):
-            metric = get_scorer(*args)
-            metric_dict[metric.name] = metric
+            if "delta_x" in self.bo_params:
+                if self.bo_params["delta_x"] < 0:
+                    raise ValueError(
+                        "Invalid value for the delta_x parameter. "
+                        f"Value should be >=0, got {self.bo_params['delta_x']}."
+                    )
+                delta_x_callback = DeltaXStopper(self.bo_params["delta_x"])
+                self._bo["callback"].append(delta_x_callback)
 
-        return metric_dict
+            if "delta_y" in self.bo_params:
+                if self.bo_params["delta_y"] < 0:
+                    raise ValueError(
+                        "Invalid value for the delta_y parameter. Value "
+                        f"should be >=0, got {self.bo_params['delta_y']}."
+                    )
+                delta_y_callback = DeltaYStopper(self.bo_params["delta_y"], n_best=5)
+                self._bo["callback"].append(delta_y_callback)
+
+            if self.bo_params.get("plot"):
+                self._bo["callback"].append(PlotCallback(self))
+
+            if "cv" in self.bo_params:
+                if self.bo_params["cv"] <= 0:
+                    raise ValueError(
+                        "Invalid value for the cv parameter. Value "
+                        f"should be >=0, got {self.bo_params['cv']}."
+                    )
+                self._bo["cv"] = self.bo_params["cv"]
+
+            if "early_stopping" in self.bo_params:
+                if self.bo_params["early_stopping"] <= 0:
+                    raise ValueError(
+                        "Invalid value for the early_stopping parameter. Value "
+                        f"should be >=0, got {self.bo_params['early_stopping']}."
+                    )
+                for model in self._models.values():
+                    if hasattr(model, "custom_fit"):
+                        model._early_stopping = self.bo_params["early_stopping"]
+
+            # Add hyperparameter dimensions to every model subclass
+            if self.bo_params.get("dimensions"):
+                for name, model in self._models.items():
+                    # If not dict, the dimensions are for all models
+                    if not isinstance(self.bo_params["dimensions"], dict):
+                        model._dimensions = lst(self.bo_params["dimensions"])
+                    else:
+                        # Dimensions for every specific model
+                        for key, value in self.bo_params["dimensions"].items():
+                            if key.lower() == name.lower() or key.lower() == "all":
+                                model._dimensions.extend(list(lst(value)))
+
+            # The remaining bo_params are added as kwargs to the optimizer
+            self._bo["kwargs"] = {
+                k: v for k, v in self.bo_params.items()
+                if k not in list(self._bo) + exc
+            }
 
     def _core_iteration(self):
         """Fit and evaluate the models in the pipeline."""
         t_init = datetime.now()  # Measure the time the whole pipeline takes
 
+        self.log("\nTraining " + "=" * 25 + " >>", 1)
+        if not self.__class__.__name__.startswith("SuccessiveHalving"):
+            self.log(f"Models: {', '.join(lst(self.models))}", 1)
+        self.log(f"Metric: {', '.join(lst(self.metric))}", 1)
+
         to_remove = []
-        for i, m in enumerate(self._models):
+        for i, m in enumerate(self._models.values()):
             model_time = datetime.now()
 
             try:  # If an error occurs, skip the model
@@ -473,17 +490,17 @@ class BaseTrainer(BaseTransformer, BasePredictor):
                     "\nException encountered while running the "
                     f"{m.name} model. Removing model from pipeline. ", 1
                 )
-                self.log("".join(traceback.format_tb(ex.__traceback__))[:-1], 5)
+                self.log("".join(traceback.format_tb(ex.__traceback__))[:-1], 3)
                 self.log(f"{type(ex).__name__}: {ex}", 1)
 
                 # Append exception to errors dictionary
-                self.errors[m.name] = ex
+                self._errors[m.name] = ex
 
                 # Add model to "garbage collector"
                 # Cannot remove immediately to maintain the iteration order
                 to_remove.append(m.name)
 
-                if self.bo_params.get("plot"):
+                if self.bo_params and self.bo_params.get("plot"):
                     PlotCallback.c += 1  # Next model
                     plt.close()  # Close the crashed plot
             finally:
@@ -491,20 +508,27 @@ class BaseTrainer(BaseTransformer, BasePredictor):
 
         delete(self, to_remove)  # Remove faulty models
 
-        # Raise an exception if all models failed
+        # If there's only one model and it failed, raise that exception
+        # If multiple models and all failed, raise RuntimeError
         if not self._models:
-            raise RuntimeError("It appears all models failed to run...")
+            if len(self.errors) == 1:
+                raise self.errors[0]
+            else:
+                raise RuntimeError(
+                    "All models failed to run. Use the errors attribute "
+                    "or the logging file to investigate the exceptions."
+                )
 
-        self.log("\n\nFinal results ========================= >>", 1)
-        self.log(f"Duration: {time_to_str(t_init)}\n" + "-" * 42, 1)
+        self.log(f"\n\nFinal results {'=' * 20} >>", 1)
+        self.log(f"Duration: {time_to_str(t_init)}\n{'-' * 37}", 1)
 
         # Get max length of the model names
-        maxlen = max([len(m.fullname) for m in self._models])
+        maxlen = max([len(m.fullname) for m in self._models.values()])
 
         # Get best score of all the models
-        best_score = max([get_best_score(m) for m in self._models])
+        best_score = max([get_best_score(m) for m in self._models.values()])
 
-        for m in self._models:
+        for m in self._models.values():
             out = f"{m.fullname:{maxlen}s} --> {m._final_output()}"
             if get_best_score(m) == best_score and len(self._models) > 1:
                 out += " !"
