@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
-"""Automated Tool for Optimized Modelling (ATOM).
-
+"""
+Automated Tool for Optimized Modelling (ATOM)
 Author: Mavs
 Description: Module containing the BaseTransformer class.
 
@@ -19,10 +19,11 @@ import multiprocessing
 from copy import deepcopy
 from typeguard import typechecked
 from typing import Union, Optional
+from sklearn.model_selection import train_test_split
 
 # Own modules
 from .utils import (
-    SEQUENCE, X_TYPES, Y_TYPES, to_df, to_series, merge,
+    SEQUENCE, X_TYPES, Y_TYPES, lst, to_df, to_series, merge,
     prepare_logger, composed, crash, method_to_log,
 )
 
@@ -158,7 +159,7 @@ class BaseTransformer:
 
         Parameters
         ----------
-        X: dict, list, tuple, np.ndarray or pd.DataFrame
+        X: dataframe-like
             Feature set with shape=(n_samples, n_features).
 
         y: int, str, sequence or None, optional (default=None)
@@ -169,10 +170,10 @@ class BaseTransformer:
 
         Returns
         -------
-        X: pd.DataFrame
+        pd.DataFrame
             Feature dataset.
 
-        y: pd.Series
+        pd.Series
             Target column corresponding to X.
 
         """
@@ -180,13 +181,16 @@ class BaseTransformer:
         # create a dataframe with one multidimensional column
         array = np.array(X)
         if array.ndim > 2 and not isinstance(array[0, 0, 0], str):
-            X = pd.DataFrame({"Multidimensional feature": [row for row in X]})
+            X = pd.DataFrame({"multidim feature": [row for row in X]})
         else:
             X = to_df(deepcopy(X))  # Make copy to not overwrite mutable arguments
 
-            # If text dataset, change the name of the column to Corpus
-            if X.shape[1] == 1 and X[X.columns[0]].dtype == "object":
-                X = X.rename(columns={X.columns[0]: "Corpus"})
+            # If text dataset, change the name of the column to corpus
+            if list(X.columns) == ["feature 1"] and X[X.columns[0]].dtype == "object":
+                X = X.rename(columns={X.columns[0]: "corpus"})
+
+            # Convert all column names to str
+            X.columns = [str(col) for col in X.columns]
 
         # Prepare target column
         if isinstance(y, (dict, *SEQUENCE)):
@@ -219,8 +223,70 @@ class BaseTransformer:
 
         return X, y
 
-    def _get_data_and_idx(self, arrays, y=-1, use_n_rows=True):
-        """Get the dataset and indices from a sequence of indexables.
+    def _set_index(self, df):
+        """Assign an index to the dataframe."""
+        target = df.columns[-1]
+
+        if self.index is True:  # True gets caught by isinstance(int)
+            return df
+        elif self.index is False:
+            df = df.reset_index(drop=True)
+        elif isinstance(self.index, int):
+            if -df.shape[1] <= self.index <= df.shape[1]:
+                df = df.set_index(df.columns[self.index], drop=True)
+            else:
+                raise ValueError(
+                    f"Invalid value for the index parameter. Value {self.index} "
+                    f"is out of range for a dataset with {df.shape[1]} columns."
+                )
+        elif isinstance(self.index, str):
+            if self.index in df:
+                df = df.set_index(self.index, drop=True)
+            else:
+                raise ValueError(
+                    "Invalid value for the index parameter. "
+                    f"Column {self.index} not found in the dataset."
+                )
+
+        if df.index.name == target:
+            raise ValueError(
+                "Invalid value for the index parameter. The index column "
+                f"can not be the same as the target column, got {target}."
+            )
+
+        return df
+
+    def _get_stratify_columns(self, df):
+        """Get columns to stratify by."""
+        # Stratification is not possible when the data can not change order
+        if self.shuffle is False or self.stratify is False:
+            return None
+        elif self.stratify is True:
+            return df.iloc[:, -1]
+        else:
+            inc = []
+            for col in lst(self.stratify):
+                if isinstance(col, int):
+                    if -df.shape[1] <= col <= df.shape[1]:
+                        inc.append(df.columns[col])
+                    else:
+                        raise ValueError(
+                            f"Invalid value for the stratify parameter. Value {col} "
+                            f"is out of range for a dataset with {df.shape[1]} columns."
+                        )
+                elif isinstance(col, str):
+                    if col in df:
+                        inc.append(col)
+                    else:
+                        raise ValueError(
+                            "Invalid value for the stratify parameter. "
+                            f"Column {col} not found in the dataset."
+                        )
+
+            return df[inc]
+
+    def _get_data(self, arrays, y=-1, use_n_rows=True):
+        """Get the data sets and indices from a sequence of indexables.
 
         Parameters
         ----------
@@ -233,50 +299,117 @@ class BaseTransformer:
         use_n_rows: bool, optional (default=True)
             Whether to use the `n_rows` parameter on the dataset.
 
+        Returns
+        -------
+        pd.DataFrame
+            Dataset containing the train and test sets.
+
+        list
+            Sizes of the train and test sets.
+
+        pd.DataFrame or None
+            Holdout data set. Returns None if not specified.
+
         """
 
-        def _no_train_test(data):
-            """Path to follow when no train and test are provided."""
-            if use_n_rows:
-                if self.n_rows > len(data):
+        def _no_data_sets(data):
+            """Path to follow when no data sets are provided."""
+            # If the index is a sequence, assign it before shuffling
+            if isinstance(self.index, SEQUENCE):
+                if len(self.index) != len(data):
                     raise ValueError(
-                        "Invalid value for the n_rows parameter. The provided"
-                        " number is larger than the number of samples, got "
-                        f"len(data)={len(data)} and n_rows={int(self.n_rows)}."
+                        "Invalid value for the index parameter. Length of "
+                        f"index ({len(self.index)}) doesn't match that of "
+                        f"the dataset ({len(data)})."
+                    )
+                data.index = self.index
+
+            if use_n_rows:
+                if not 0 < self.n_rows <= len(data):
+                    raise ValueError(
+                        "Invalid value for the n_rows parameter. Value "
+                        f"should lie between 0 and len(X), got {self.n_rows}."
                     )
 
                 # Select subset of the data
-                n = self.n_rows if self.n_rows > 1 else len(data) * self.n_rows
-                if self.shuffle:
-                    data = data.sample(n=int(n), random_state=self.random_state)
+                if self.n_rows > 1:
+                    data = data.iloc[:int(self.n_rows), :]
                 else:
-                    data = data.iloc[:int(n), :]
+                    data = data.iloc[:int(len(data) * self.n_rows), :]
 
-            data = data.reset_index(drop=True)
-
-            if len(data) < 2:
+            if len(data) < 5:
                 raise ValueError(
-                    "Invalid value for the n_rows parameter, got "
-                    f"{self.n_rows}. Length of dataset can not be <2."
+                    "Invalid value for the n_rows parameter. The "
+                    f"length of the dataset can't be <5, got {self.n_rows}."
                 )
 
-            if self.test_size <= 0 or self.test_size >= len(data):
+            if not 0 < self.test_size < len(data):
                 raise ValueError(
                     "Invalid value for the test_size parameter. Value "
                     f"should lie between 0 and len(X), got {self.test_size}."
                 )
 
-            # Define train and test indices
+            # Define test set size
             if self.test_size < 1:
-                test_idx = int(self.test_size * len(data))
+                test_size = max(1, int(self.test_size * len(data)))
             else:
-                test_idx = self.test_size
-            idx = [len(data) - test_idx, test_idx]
+                test_size = self.test_size
 
-            return data, idx
+            # Define holdout set size
+            if self.holdout_size:
+                if self.holdout_size < 1:
+                    holdout_size = max(1, int(self.holdout_size * len(data)))
+                else:
+                    holdout_size = self.holdout_size
+
+                if not 0 <= holdout_size <= len(data) - test_size:
+                    raise ValueError(
+                        "Invalid value for the holdout_size parameter. "
+                        "Value should lie between 0 and len(X) - len(test), "
+                        f"got {self.holdout_size}."
+                    )
+
+                data, holdout = train_test_split(
+                    data,
+                    test_size=holdout_size,
+                    random_state=self.random_state,
+                    shuffle=self.shuffle,
+                    stratify=self._get_stratify_columns(data),
+                )
+                holdout = self._set_index(holdout)
+            else:
+                holdout = None
+
+            train, test = train_test_split(
+                data,
+                test_size=test_size,
+                random_state=self.random_state,
+                shuffle=self.shuffle,
+                stratify=self._get_stratify_columns(data),
+            )
+            data = self._set_index(pd.concat([train, test]))
+
+            return data, [data.index[:-test_size], data.index[-test_size:]], holdout
+
+        def _has_data_sets(train, test, holdout=None):
+            """Path to follow when data sets are provided."""
+            # If the index is a sequence, assign it before shuffling
+            if isinstance(self.index, SEQUENCE):
+                len_data = len(train) + len(test)
+                if holdout is not None:
+                    len_data += len(holdout)
+
+                if len(self.index) != len_data:
+                    raise ValueError(
+                        "Invalid value for the index parameter. Length of "
+                        f"index ({len(self.index)}) doesn't match that of "
+                        f"the data sets ({len_data})."
+                    )
+                train.index = self.index[:len(train)]
+                test.index = self.index[len(train):len(train) + len(test)]
+                if holdout is not None:
+                    holdout.index = self.index[-len(holdout):]
 
-        def _has_train_test(train, test):
-            """Path to follow when train and test are provided."""
             # Skip the n_rows step if not called from atom
             if hasattr(self, "n_rows") and use_n_rows:
                 # Select same subsample of train and test set
@@ -289,61 +422,104 @@ class BaseTransformer:
                     else:
                         train = train.iloc[:n_train, :]
                         test = test.iloc[:n_test, :]
+
+                    if holdout is not None:
+                        n_holdout = int(len(holdout) * self.n_rows)
+                        if self.shuffle:
+                            holdout = holdout.sample(
+                                n=n_holdout,
+                                random_state=self.random_state,
+                            )
+                        else:
+                            holdout = holdout.iloc[:n_holdout, :]
+
                 else:
                     raise ValueError(
                         "Invalid value for the n_rows parameter. Value has "
                         "to be <1 when a train and test set are provided."
                     )
 
-            data = pd.concat([train, test]).reset_index(drop=True)
-            idx = [len(train), len(test)]
+            if not train.columns.equals(test.columns):
+                raise ValueError("The train and test set do not have the same columns.")
 
-            return data, idx
+            if holdout is not None:
+                if not train.columns.equals(holdout.columns):
+                    raise ValueError(
+                        "The holdout set does not have the "
+                        "same columns as the train and test set."
+                    )
+                holdout = self._set_index(holdout)
+
+            data = self._set_index(pd.concat([train, test]))
+            idx = [data.index[:len(train)], data.index[-len(test):]]
+
+            return data, idx, holdout
 
         # Process input arrays ===================================== >>
 
         if len(arrays) == 0:
-            if self.branch.data is None:
+            if self.branch._data is None:
                 raise ValueError(
                     "The data arrays are empty! Provide the data to run the pipeline "
                     "successfully. See the documentation for the allowed formats."
                 )
             else:
-                return self.branch.data, self.branch.idx
+                return self.branch._data, self.branch._idx, self.holdout
 
         elif len(arrays) == 1:
             # arrays=(X,)
             data = merge(*self._prepare_input(arrays[0], y=y))
-            data, idx = _no_train_test(data)
+            data, idx, holdout = _no_data_sets(data)
 
         elif len(arrays) == 2:
-            if len(arrays[0]) == len(arrays[1]) == 2:
+            if isinstance(arrays[0], tuple) and len(arrays[0]) == len(arrays[1]) == 2:
                 # arrays=((X_train, y_train), (X_test, y_test))
                 train = merge(*self._prepare_input(arrays[0][0], arrays[0][1]))
                 test = merge(*self._prepare_input(arrays[1][0], arrays[1][1]))
-                data, idx = _has_train_test(train, test)
+                data, idx, holdout = _has_data_sets(train, test)
             elif isinstance(arrays[1], (int, str)) or np.array(arrays[1]).ndim == 1:
                 # arrays=(X, y)
                 data = merge(*self._prepare_input(arrays[0], arrays[1]))
-                data, idx = _no_train_test(data)
+                data, idx, holdout = _no_data_sets(data)
             else:
                 # arrays=(train, test)
                 train = merge(*self._prepare_input(arrays[0], y=y))
                 test = merge(*self._prepare_input(arrays[1], y=y))
-                data, idx = _has_train_test(train, test)
+                data, idx, holdout = _has_data_sets(train, test)
+
+        elif len(arrays) == 3:
+            if len(arrays[0]) == len(arrays[1]) == len(arrays[2]) == 2:
+                # arrays=((X_train, y_train), (X_test, y_test), (X_holdout, y_holdout))
+                train = merge(*self._prepare_input(arrays[0][0], arrays[0][1]))
+                test = merge(*self._prepare_input(arrays[1][0], arrays[1][1]))
+                holdout = merge(*self._prepare_input(arrays[2][0], arrays[2][1]))
+                data, idx, holdout = _has_data_sets(train, test, holdout)
+            else:
+                # arrays=(train, test, holdout)
+                train = merge(*self._prepare_input(arrays[0], y=y))
+                test = merge(*self._prepare_input(arrays[1], y=y))
+                holdout = merge(*self._prepare_input(arrays[2], y=y))
+                data, idx, holdout = _has_data_sets(train, test, holdout)
 
         elif len(arrays) == 4:
             # arrays=(X_train, X_test, y_train, y_test)
             train = merge(*self._prepare_input(arrays[0], arrays[2]))
             test = merge(*self._prepare_input(arrays[1], arrays[3]))
-            data, idx = _has_train_test(train, test)
+            data, idx, holdout = _has_data_sets(train, test)
+
+        elif len(arrays) == 6:
+            # arrays=(X_train, X_test, X_holdout, y_train, y_test, y_holdout)
+            train = merge(*self._prepare_input(arrays[0], arrays[3]))
+            test = merge(*self._prepare_input(arrays[1], arrays[4]))
+            holdout = merge(*self._prepare_input(arrays[2], arrays[5]))
+            data, idx, holdout = _has_data_sets(train, test, holdout)
 
         else:
             raise ValueError(
                 "Invalid data arrays. See the documentation for the allowed formats."
             )
 
-        return data, idx
+        return data, idx, holdout
 
     @composed(crash, typechecked)
     def log(self, msg: Union[int, float, str], level: int = 0):
@@ -382,10 +558,11 @@ class BaseTransformer:
 
         """
         if not save_data and hasattr(self, "dataset"):
-            data = {}  # Store the data to reattach later
+            data = {"holdout": deepcopy(self.holdout)}  # Store data to reattach later
+            self.holdout = None
             for key, value in self._branches.items():
-                data[key] = deepcopy(value.data)
-                value.data = None
+                data[key] = deepcopy(value._data)
+                value._data = None
 
         if filename.endswith("auto"):
             filename = filename.replace("auto", self.__class__.__name__)
@@ -395,7 +572,8 @@ class BaseTransformer:
 
         # Restore the data to the attributes
         if not save_data and hasattr(self, "dataset"):
+            self.holdout = data["holdout"]
             for key, value in self._branches.items():
-                value.data = data[key]
+                value._data = data[key]
 
-        self.log(self.__class__.__name__ + " saved successfully!", 1)
+        self.log(f"{self.__class__.__name__} successfully saved.", 1)
