@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
-"""Automated Tool for Optimized Modelling (ATOM).
-
+"""
+Automated Tool for Optimized Modelling (ATOM)
 Author: Mavs
 Description: Module containing the training classes.
 
@@ -17,7 +17,7 @@ from sklearn.base import BaseEstimator
 
 # Own modules
 from .basetrainer import BaseTrainer
-from .plots import BaseModelPlotter, SuccessiveHalvingPlotter, TrainSizingPlotter
+from .plots import BaseModelPlotter
 from .utils import (
     SEQUENCE_TYPES, lst, get_best_score, infer_task, composed,
     method_to_log, crash, CustomDict,
@@ -53,24 +53,20 @@ class Direct(BaseEstimator, BaseTrainer, BaseModelPlotter):
         Parameters
         ----------
         *arrays: sequence of indexables
-            Training set and test set. Allowed input formats are:
+            Training set and test set. Allowed formats are:
                 - train, test
                 - X_train, X_test, y_train, y_test
                 - (X_train, y_train), (X_test, y_test)
 
         """
-        self.branch.data, self.branch.idx = self._get_data_and_idx(arrays)
+        self.branch._data, self.branch._idx, self.holdout = self._get_data(arrays)
         self.task = infer_task(self.y_train, goal=self.goal)
-        self._check_parameters()
-
-        self.log("\nTraining ===================================== >>", 1)
-        self.log(f"Models: {', '.join(lst(self.models))}", 1)
-        self.log(f"Metric: {', '.join(lst(self.metric))}", 1)
+        self._prepare_parameters()
 
         self._core_iteration()
 
 
-class SuccessiveHalving(BaseEstimator, BaseTrainer, SuccessiveHalvingPlotter):
+class SuccessiveHalving(BaseEstimator, BaseTrainer, BaseModelPlotter):
     """Successive halving training approach.
 
     The successive halving technique is a bandit-based algorithm that
@@ -110,15 +106,15 @@ class SuccessiveHalving(BaseEstimator, BaseTrainer, SuccessiveHalvingPlotter):
         Parameters
         ----------
         *arrays: sequence of indexables
-            Training set and test set. Allowed input formats are:
+            Training set and test set. Allowed formats are:
                 - train, test
                 - X_train, X_test, y_train, y_test
                 - (X_train, y_train), (X_test, y_test)
 
         """
-        self.branch.data, self.branch.idx = self._get_data_and_idx(arrays)
+        self.branch._data, self.branch._idx, self.holdout = self._get_data(arrays)
         self.task = infer_task(self.y_train, goal=self.goal)
-        self._check_parameters()
+        self._prepare_parameters()
 
         if self.skip_runs < 0:
             raise ValueError(
@@ -132,17 +128,14 @@ class SuccessiveHalving(BaseEstimator, BaseTrainer, SuccessiveHalvingPlotter):
                 f"and skip_runs={self.skip_runs}."
             )
 
-        self.log("\nTraining ===================================== >>", 1)
-        self.log(f"Metric: {', '.join(lst(self.metric))}", 1)
-
         run = 0
         models = CustomDict()
         og_models = {k: copy(v) for k, v in self._models.items()}
         while len(self._models) > 2 ** self.skip_runs - 1:
             # Create the new set of models for the run
-            for m in self._models:
+            for m in self._models.values():
                 m.name += str(len(self._models))
-                m._pred_attrs = [None] * 10  # Avoid shallow copy
+                m._pred = [None] * 15  # Avoid shallow copy
                 m._train_idx = len(self.train) // len(self._models)
 
             # Print stats for this subset of the data
@@ -153,16 +146,16 @@ class SuccessiveHalving(BaseEstimator, BaseTrainer, SuccessiveHalvingPlotter):
             self.log(f"Size of test set: {len(self.test)}", 1)
 
             self._core_iteration()
-            models.update({m.name: m for m in self._models})
+            models.update({m.name: m for m in self._models.values()})
 
             # Select next models for halving
             best = pd.Series(
-                data=[get_best_score(m) for m in self._models],
-                index=[m.name for m in self._models],
+                data=[get_best_score(m) for m in self._models.values()],
+                index=[m.name for m in self._models.values()],
             ).nlargest(n=len(self._models) // 2, keep="first")
-            acronyms = [m.acronym for m in self._models if m.name in best.index]
+            names = [m.acronym for m in self._models.values() if m.name in best.index]
             self._models = CustomDict(
-                {k: copy(v) for k, v in og_models.items() if v.acronym in acronyms}
+                {k: copy(v) for k, v in og_models.items() if v.acronym in names}
             )
 
             run += 1
@@ -170,7 +163,7 @@ class SuccessiveHalving(BaseEstimator, BaseTrainer, SuccessiveHalvingPlotter):
         self._models = models  # Restore all models
 
 
-class TrainSizing(BaseEstimator, BaseTrainer, TrainSizingPlotter):
+class TrainSizing(BaseEstimator, BaseTrainer, BaseModelPlotter):
     """Train Sizing training approach.
 
     When training models, there is usually a trade-off between model
@@ -212,19 +205,15 @@ class TrainSizing(BaseEstimator, BaseTrainer, TrainSizingPlotter):
         Parameters
         ----------
         *arrays: sequence of indexables
-            Training set and test set. Allowed input formats are:
+            Training set and test set. Allowed formats are:
                 - train, test
                 - X_train, X_test, y_train, y_test
                 - (X_train, y_train), (X_test, y_test)
 
         """
-        self.branch.data, self.branch.idx = self._get_data_and_idx(arrays)
+        self.branch._data, self.branch._idx, self.holdout = self._get_data(arrays)
         self.task = infer_task(self.y_train, goal=self.goal)
-        self._check_parameters()
-
-        self.log("\nTraining ===================================== >>", 1)
-        self.log(f"Models: {', '.join(lst(self.models))}", 1)
-        self.log(f"Metric: {', '.join(lst(self.metric))}", 1)
+        self._prepare_parameters()
 
         # Convert integer train_sizes to sequence
         if isinstance(self.train_sizes, int):
@@ -236,24 +225,24 @@ class TrainSizing(BaseEstimator, BaseTrainer, TrainSizingPlotter):
             # Select fraction of data to use in this run
             if size <= 1:
                 frac = round(size, 2)
-                train_idx = int(size * self.branch.idx[0])
+                train_idx = int(size * len(self.branch.train))
             else:
-                frac = round(size / self.branch.idx[0], 2)
+                frac = round(size / len(self.branch.train), 2)
                 train_idx = size
 
-            for m in self._models:
+            for m in self._models.values():
                 m.name += str(frac).replace(".", "")  # Add frac to the name
-                m._pred_attrs = [None] * 10  # Avoid shallow copy
+                m._pred = [None] * 15  # Avoid shallow copy
                 m._train_idx = train_idx
 
             # Print stats for this subset of the data
-            p = round(train_idx * 100.0 / self.branch.idx[0])
+            p = round(train_idx * 100.0 / len(self.branch.train))
             self.log(f"\n\nRun: {run} {'='*32} >>", 1)
             self.log(f"Size of training set: {train_idx} ({p}%)", 1)
             self.log(f"Size of test set: {len(self.test)}", 1)
 
             self._core_iteration()
-            models.update({m.name.lower(): m for m in self._models})
+            models.update({m.name.lower(): m for m in self._models.values()})
 
             # Create next models for sizing
             self._models = CustomDict({k: copy(v) for k, v in og_models.items()})
@@ -284,7 +273,7 @@ class DirectClassifier(Direct):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "classification"
+        self.goal = "class"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             n_calls, n_initial_points, est_params, bo_params, n_bootstrap,
@@ -315,7 +304,7 @@ class DirectRegressor(Direct):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "regression"
+        self.goal = "reg"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             n_calls, n_initial_points, est_params, bo_params, n_bootstrap,
@@ -347,7 +336,7 @@ class SuccessiveHalvingClassifier(SuccessiveHalving):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "classification"
+        self.goal = "class"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             skip_runs, n_calls, n_initial_points, est_params, bo_params,
@@ -380,7 +369,7 @@ class SuccessiveHalvingRegressor(SuccessiveHalving):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "regression"
+        self.goal = "reg"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             skip_runs, n_calls, n_initial_points, est_params, bo_params,
@@ -413,7 +402,7 @@ class TrainSizingClassifier(TrainSizing):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "classification"
+        self.goal = "class"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             train_sizes, n_calls, n_initial_points, est_params, bo_params,
@@ -446,7 +435,7 @@ class TrainSizingRegressor(TrainSizing):
         experiment: Optional[str] = None,
         random_state: Optional[int] = None,
     ):
-        self.goal = "regression"
+        self.goal = "reg"
         super().__init__(
             models, metric, greater_is_better, needs_proba, needs_threshold,
             train_sizes, n_calls, n_initial_points, est_params, bo_params,
