@@ -9,6 +9,8 @@ Description: Unit tests for atom.py
 
 # Standard packages
 import glob
+
+import pandas as pd
 import pytest
 import numpy as np
 from unittest.mock import patch
@@ -20,49 +22,26 @@ from sklearn.linear_model import LassoLarsCV
 from sklearn.ensemble import RandomForestClassifier
 from sklearn.pipeline import Pipeline
 from sklearn.neural_network import MLPClassifier
-from sklearn.preprocessing import LabelEncoder, StandardScaler, RobustScaler
+from sklearn.preprocessing import (
+    LabelEncoder,
+    OneHotEncoder,
+    StandardScaler,
+    RobustScaler,
+)
 
 # Own modules
 from atom import ATOMClassifier, ATOMRegressor
 from atom.data_cleaning import Scaler, Pruner
 from atom.utils import check_scaling
 from .utils import (
-    FILE_DIR, X_bin, y_bin, X_class, y_class, X_reg, y_reg, X_text,
-    y_text, X10, X10_nan, X10_str, X10_str2, X10_dt, y10, y10_str,
+    FILE_DIR, X_bin, y_bin, X_class, y_class, X_reg, y_reg, X_sparse,
+    X_text, X10, X10_nan, X10_str, X10_str2, X10_dt, y10, y10_str,
     y10_sn, X20_out,
 )
 
 
 # Test __init__ ==================================================== >>
 
-def test_dtypes_shrinkage():
-    """Assert that the dtypes are optimized to save memory."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    assert not atom.dtypes.equals(X_bin.dtypes)
-
-
-def test_minimal_dtypes_unchanged():
-    """Assert that the minimal dtypes are left unchanged."""
-    X = X_bin.astype("float32")
-    atom = ATOMClassifier(X, y_bin, random_state=1)
-    assert atom.X.dtypes.equals(X.dtypes)
-
-
-def test_dtypes_cat_or_object():
-    """Assert that categorical columns change dtype if necessary."""
-    atom = ATOMClassifier(X10_str, y10, random_state=1)
-    assert atom.dtypes[2].name != "category"
-
-    atom = ATOMClassifier(X10_str2, y10, random_state=1)
-    assert atom.dtypes[2].name == "category"
-
-
-def test_test_size_attribute():
-    """Assert that the _test_size attribute is created."""
-    atom = ATOMClassifier(X_bin, y_bin, test_size=0.3, n_jobs=2, random_state=1)
-    assert atom._test_size == len(atom.test) / len(atom.dataset)
-
-
 def test_task_assignment():
     """Assert that the correct task is assigned."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
@@ -81,18 +60,6 @@ def test_raise_one_target_value():
     pytest.raises(ValueError, ATOMClassifier, X_bin, y, random_state=1)
 
 
-def test_mapping_assignment():
-    """Assert that the mapping attribute is created."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    assert atom.mapping == {"0": 0, "1": 1}
-
-
-def test_mapping_with_nans():
-    """Assert that the mapping attribute is created when str and nans are mixed."""
-    atom = ATOMClassifier(X10, y10_sn, random_state=1)
-    assert atom.mapping == {"n": "n", "nan": np.NaN, "y": "y"}
-
-
 # Test magic methods =============================================== >>
 
 def test_repr():
@@ -100,8 +67,8 @@ def test_repr():
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.scale()
     assert "Branches: master" in str(atom)
-    atom.branch = "branch_2"
-    assert "Branches:\n   >>> master\n   >>> branch_2 !" in str(atom)
+    atom.branch = "b2"
+    assert "Branches:\n   >>> master\n   >>> b2 !" in str(atom)
 
 
 def test_iter():
@@ -112,87 +79,71 @@ def test_iter():
     assert [item for item in atom][1] == atom.pipeline[1]
 
 
-def test_contains():
-    """Assert that we can test if atom contains a column."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    assert "mean radius" in atom
-
-
-def test_len():
-    """Assert that the length of atom is the length of the pipeline."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.clean()
-    assert len(atom) == 1
-
-
-def test_getitem():
-    """Assert that atom is subscriptable."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.clean()
-    atom.impute()
-    assert atom[1].__class__.__name__ == "Imputer"
-    assert atom["mean radius"].equals(atom.dataset["mean radius"])
-    with pytest.raises(TypeError, match=r".*subscriptable with types.*"):
-        print(atom[2.3])
-
-
 # Test utility properties =========================================== >>
 
-def test_branch_setter_empty():
-    """Assert that an error is raised when the name is empty."""
+def test_branch_same():
+    """Assert that we can stay on the same branch."""
     atom = ATOMClassifier(X10, y10, random_state=1)
-    with pytest.raises(ValueError, match=r".*Can't create a branch.*"):
+    atom.branch = "master"
+    assert atom.branch.name == "master"
+
+
+def test_branch_change():
+    """Assert that we can change to another branch."""
+    atom = ATOMClassifier(X10, y10, random_state=1)
+    atom.branch = "b2"
+    atom.clean()
+    atom.branch = "master"
+    assert atom.pipeline.empty  # Has no Cleaner
+
+
+def test_branch_empty():
+    """Assert that an error is raised when name is empty."""
+    atom = ATOMClassifier(X10, y10, random_state=1)
+    with pytest.raises(ValueError, match=r".*have an empty name.*"):
         atom.branch = ""
 
 
+def test_branch_existing_name():
+    """Assert that an error is raised when the name already exists."""
+    atom = ATOMClassifier(X10, y10, random_state=1)
+    atom.branch = "b2"
+    with pytest.raises(ValueError, match=r".*already exists.*"):
+        atom.branch = "b2_from_master"
+
+
 def test_branch_model_acronym():
     """Assert that an error is raised when the name is a models' acronym."""
     atom = ATOMClassifier(X10, y10, random_state=1)
-    with pytest.raises(ValueError, match=r".*acronym of model.*"):
+    with pytest.raises(ValueError, match=r".*model's acronym.*"):
         atom.branch = "Lda"
 
 
-def test_branch_restricted_name():
-    """Assert that an error is raised when the name is restricted."""
+def test_branch_unknown_parent():
+    """Assert that an error is raised when the parent doesn't exist."""
     atom = ATOMClassifier(X10, y10, random_state=1)
-    with pytest.raises(ValueError, match=r".*This name is reserved.*"):
-        atom.branch = "og"
+    with pytest.raises(ValueError, match=r".*does not exist.*"):
+        atom.branch = "b2_from_invalid"
 
 
-def test_branch_setter_change():
-    """Assert that we can change to an old branch."""
-    atom = ATOMClassifier(X10, y10, random_state=1)
-    atom.branch = "branch_2"
-    atom.clean()
-    atom.branch = "master"
-    assert atom.pipeline.empty  # Has no clean estimator
-
-
-def test_branch_setter_new():
+def test_branch_new():
     """Assert that we can create a new branch."""
     atom = ATOMClassifier(X10, y10, random_state=1)
     atom.clean()
-    atom.branch = "branch_2"
-    assert list(atom._branches.keys()) == ["og", "master", "branch_2"]
+    atom.branch = "b2"
+    assert list(atom._branches) == ["og", "master", "b2"]
 
 
-def test_branch_setter_from_valid():
+def test_branch_from_valid():
     """Assert that we can create a new branch, not from the current one."""
     atom = ATOMClassifier(X10_nan, y10, random_state=1)
-    atom.branch = "branch_2"
+    atom.branch = "b2"
     atom.impute()
-    atom.branch = "branch_3_from_master"
-    assert atom.branch.name == "branch_3"
+    atom.branch = "b3_from_master"
+    assert atom.branch.name == "b3"
     assert atom.n_nans > 0
 
 
-def test_branch_setter_from_invalid():
-    """Assert that an error is raised when the from branch doesn't exist."""
-    atom = ATOMClassifier(X10, y10, random_state=1)
-    with pytest.raises(ValueError, match=r".*branch to split from does not exist.*"):
-        atom.branch = "new_branch_from_invalid"
-
-
 def test_scaled():
     """Assert that scaled returns if the dataset is scaled."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
@@ -222,19 +173,19 @@ def test_n_nans():
 def test_numerical():
     """Assert that numerical returns the names of the numerical columns."""
     atom = ATOMClassifier(X10_str, y10, random_state=1)
-    assert atom.numerical == ["Feature 1", "Feature 2"]
+    assert atom.numerical == ["feature 1", "feature 2", "feature 4"]
 
 
 def test_n_numerical():
     """Assert that n_categorical returns the number of numerical columns."""
     atom = ATOMClassifier(X10_str, y10, random_state=1)
-    assert atom.n_numerical == 2
+    assert atom.n_numerical == 3
 
 
 def test_categorical():
     """Assert that categorical returns the names of categorical columns."""
     atom = ATOMClassifier(X10_str, y10, random_state=1)
-    assert atom.categorical == ["Feature 3"]
+    assert atom.categorical == ["feature 3"]
 
 
 def test_n_categorical():
@@ -255,13 +206,13 @@ def test_n_outliers():
     assert atom.n_outliers == 2
 
 
-def test_classes_property():
+def test_classes():
     """Assert that the classes property returns a df of the classes in y."""
     atom = ATOMClassifier(X_class, y_class, random_state=1)
     assert list(atom.classes.index) == [0, 1, 2]
 
 
-def test_n_classes_property():
+def test_n_classes():
     """Assert that the n_classes property returns the number of classes."""
     atom = ATOMClassifier(X_class, y_class, random_state=1)
     assert atom.n_classes == 3
@@ -269,88 +220,6 @@ def test_n_classes_property():
 
 # Test utility methods ============================================= >>
 
-def test_status():
-    """Assert that the status method prints an overview of the instance."""
-    atom = ATOMClassifier(X_class, y_class, random_state=1)
-    atom.status()
-
-
-def test_reset():
-    """Assert that the reset method deletes models and branches."""
-    atom = ATOMClassifier(X_class, y_class, random_state=1)
-    atom.branch = "2"
-    atom.run("LR")
-    atom.reset()
-    assert not atom.models and len(atom._branches) == 2
-    assert atom.dataset.equals(atom.og.dataset)
-
-
-@pytest.mark.parametrize("column", ["Feature 1", 1])
-def test_distribution(column):
-    """Assert that the distribution method and file are created."""
-    atom = ATOMClassifier(X10_str, y10, random_state=1)
-    pytest.raises(ValueError, atom.distribution, column="Feature 3")
-    df = atom.distribution(column=column)
-    assert len(df) == 11
-
-
-@patch("pandas_profiling.ProfileReport")
-def test_report(cls):
-    """Assert that the report method and file are created."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.report(n_rows=10, filename="report")
-    cls.return_value.to_file.assert_called_once_with("report.html")
-
-
-def test_transform_method():
-    """ Assert that the transform method works as intended."""
-    atom = ATOMClassifier(X10_str, y10, random_state=1)
-    atom.encode(max_onehot=None)
-    assert atom.transform(X10_str)["Feature 3"].dtype.kind in "ifu"
-
-
-def test_pipeline_parameter_None():
-    """Assert that only some transformers are used."""
-    atom = ATOMClassifier(X10_nan, y10, random_state=1)
-    atom.impute(strat_num="median")
-    atom.prune(max_sigma=2)
-    X = atom.transform(X10_nan, pipeline=None)  # Only use imputer
-    assert len(X) == 10
-
-
-def test_pipeline_parameter_True():
-    """Assert that all transformers are used."""
-    atom = ATOMClassifier(X10_nan, y10, random_state=1)
-    atom.impute(strat_num="median")
-    atom.prune(max_sigma=2)
-    X = atom.transform(X10_nan, pipeline=True)  # Use both transformers
-    assert len(X) < 10
-
-
-def test_pipeline_parameter_False():
-    """Assert that no transformers used."""
-    atom = ATOMClassifier(X10_nan, y10, random_state=1)
-    atom.impute(strat_num="median")
-    atom.prune(max_sigma=2)
-    X = atom.transform(X10_nan, pipeline=False)  # Use None
-    assert isinstance(X, list)  # X is unchanged
-
-
-def test_verbose_raises_when_invalid():
-    """Assert an error is raised for an invalid value of verbose."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.clean()
-    pytest.raises(ValueError, atom.transform, X_bin, verbose=3)
-
-
-def test_transform_with_y():
-    """Assert that the transform method works when y is provided."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.prune(strategy="iforest", include_target=True)
-    X, y = atom.transform(X_bin, y_bin, pipeline=[0])
-    assert len(y) < len(y_bin)
-
-
 @patch("tpot.TPOTClassifier")
 def test_automl_classification(cls):
     """Assert that the automl method works for classification tasks."""
@@ -365,6 +234,7 @@ def test_automl_classification(cls):
 
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.run("Tree", metric="accuracy")
+    atom.branch = "automl"  # Change branch since new pipeline
     atom.automl()
     cls.assert_called_with(
         n_jobs=1,
@@ -372,7 +242,7 @@ def test_automl_classification(cls):
         scoring=atom._metric[0],  # Called using atom's metric
         verbosity=0,
     )
-    assert len(atom) == 2
+    assert len(atom.pipeline) == 2
     assert atom.models == ["Tree", "MLP"]
 
 
@@ -397,30 +267,23 @@ def test_automl_invalid_scoring():
     """Assert that an error is raised when the provided scoring is invalid."""
     atom = ATOMRegressor(X_reg, y_reg, random_state=1)
     atom.run("Tree", metric="mse")
-    pytest.raises(ValueError, atom.automl, scoring="r2")
+    with pytest.raises(ValueError, match=r".*scoring parameter.*"):
+        atom.automl(scoring="r2")
 
 
-def test_save_data():
-    """Assert that the dataset is saved to a csv file."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.save_data(FILE_DIR + "auto")
-    assert glob.glob(FILE_DIR + "ATOMClassifier_dataset.csv")
+@pytest.mark.parametrize("distributions", [None, "norm", ["norm", "pearson3"]])
+@pytest.mark.parametrize("columns", ["feature 1", 1, None])
+def test_distribution(distributions, columns):
+    """Assert that the distribution method and file are created."""
+    atom = ATOMClassifier(X10_str, y10, random_state=1)
+    df = atom.distribution(distributions=distributions, columns=columns)
+    assert isinstance(df, pd.DataFrame)
 
 
-@pytest.mark.parametrize("pl", [(None, 1), (True, 2), ([0], 1)])
-def test_export_pipeline(pl):
-    """Assert that we can export the pipeline."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.clean()
-    atom.prune()
-    assert len(atom.export_pipeline(pipeline=pl[0])) == pl[1]
-
-
-def test_export_empty_pipeline():
+def test_export_pipeline_empty():
     """Assert that an error is raised when the pipeline is empty."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.clean()
-    pytest.raises(ValueError, atom.export_pipeline, pipeline=False)
+    pytest.raises(ValueError, atom.export_pipeline)
 
 
 def test_export_pipeline_verbose():
@@ -431,14 +294,22 @@ def test_export_pipeline_verbose():
     assert atom.export_pipeline("LGB", verbose=2)[0].verbose == 2
 
 
-def test_export_same_transformer():
+def test_export_pipeline_same_transformer():
     """Assert that two same transformers get different names."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.clean()
     atom.clean()
     atom.clean()
     pl = atom.export_pipeline()
-    assert list(pl.named_steps.keys()) == ["cleaner", "cleaner2", "cleaner3"]
+    assert list(pl.named_steps) == ["cleaner", "cleaner2", "cleaner3"]
+
+
+def test_export_pipeline_invalid_model():
+    """Assert that an error is raised when model is from other branch."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.run("GNB")
+    atom.branch = "b2"
+    pytest.raises(ValueError, atom.export_pipeline, model="gnb")
 
 
 def test_export_pipeline_scaler():
@@ -449,15 +320,158 @@ def test_export_pipeline_scaler():
     assert isinstance(atom.export_pipeline("LGB")[0], Scaler)
 
 
-# Test base transformers =========================================== >>
-
-def test_params_to_method():
-    """Assert that atom's parameters are passed to the method."""
-    atom = ATOMClassifier(X_bin, y_bin, verbose=1, random_state=1)
+@patch("tempfile.gettempdir")
+def test_export_pipeline_memory(func):
+    """Assert that memory is True triggers a temp dir."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.scale()
-    assert atom.pipeline[0].verbose == 1
+    atom.export_pipeline(memory=True)
+    func.assert_called_once()
 
 
+@patch("pandas_profiling.ProfileReport")
+def test_report(cls):
+    """Assert that the report method and file are created."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.report(n_rows=10, filename="report")
+    cls.return_value.to_file.assert_called_once_with("report.html")
+
+
+def test_reset():
+    """Assert that the reset method deletes models and branches."""
+    atom = ATOMClassifier(X10_str, y10, random_state=1)
+    atom.scale()
+    atom.branch = "2"
+    atom.encode()
+    atom.run("LR")
+    atom.reset()
+    assert not atom.models and len(atom._branches) == 1
+    assert atom["feature 3"].dtype.name == "object"  # Is reset back to str
+
+
+def test_save_data():
+    """Assert that the dataset is saved to a csv file."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.save_data(FILE_DIR + "auto")
+    assert glob.glob(FILE_DIR + "ATOMClassifier_dataset.csv")
+
+
+def test_shrink_dtypes_excluded():
+    """Assert that some dtypes are excluded from changing."""
+    atom = ATOMClassifier(X10_str2, y10, random_state=1)
+    assert atom.dtypes[3].name == "bool"
+    atom.shrink()
+    assert atom.dtypes[3].name == "bool"
+
+
+def test_shrink_obj2cat():
+    """Assert that the obj2cat parameter works as intended."""
+    atom = ATOMClassifier(X10_str2, y10, random_state=1)
+    atom.shrink(obj2cat=False)
+    assert atom.dtypes[2].name == "object"
+
+    atom.shrink()
+    assert atom.dtypes[2].name == "category"
+
+
+def test_shrink_int2uint():
+    """Assert that the int2uint parameter works as intended."""
+    atom = ATOMClassifier(X10_str2, y10, random_state=1)
+    assert atom.dtypes[0].name == "int64"
+    atom.shrink()
+    assert atom.dtypes[0].name == "int8"
+
+    assert atom.dtypes[0].name == "int8"
+    atom.shrink(int2uint=True)
+    assert atom.dtypes[0].name == "uint8"
+
+
+def test_shrink_sparse_arrays():
+    """Assert that sparse arrays are also transformed."""
+    atom = ATOMClassifier(X_sparse, y10, random_state=1)
+    assert atom.dtypes[0].name == "Sparse[int64, 0]"
+    atom.shrink()
+    assert atom.dtypes[0].name == "Sparse[int8, 0]"
+
+
+def test_shrink_dtypes_unchanged():
+    """Assert that optimal dtypes are left unchanged."""
+    atom = ATOMClassifier(X_bin.astype("float32"), y_bin, random_state=1)
+    assert atom.dtypes[3].name == "float32"
+    atom.shrink()
+    assert atom.dtypes[3].name == "float32"
+
+
+def test_shrink_dense2sparse():
+    """Assert that the dataset can be converted to sparse."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    assert atom.dtypes[0].name == "float64"
+    atom.shrink(dense2sparse=True)
+    assert atom.dtypes[0].name.startswith("Sparse[float32")
+
+
+def test_shrink_exclude_columns():
+    """Assert that columns can be excluded."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    assert atom.dtypes[0].name == "float64"
+    assert atom.dtypes[-1].name != "int8"
+    atom.shrink(columns=-1)
+    assert atom.dtypes[0].name == "float64"
+    assert atom.dtypes[-1].name == "int8"
+
+
+def test_stats_mixed_sparse_dense():
+    """Assert that stats show new information for mixed datasets."""
+    atom = ATOMClassifier(X_sparse, y10, random_state=1)
+    atom.apply(lambda x: 1, columns="new_column")
+    atom.stats()
+
+
+def test_status():
+    """Assert that the status method prints an overview of the instance."""
+    atom = ATOMClassifier(X_class, y_class, random_state=1)
+    atom.status()
+
+
+def test_transform_method():
+    """ Assert that the transform method works as intended."""
+    atom = ATOMClassifier(X10_str, y10, random_state=1)
+    atom.encode(max_onehot=None)
+    assert atom.transform(X10_str)["feature 3"].dtype.kind in "ifu"
+
+
+def test_transform_not_train_only():
+    """Assert that train_only transformers are not used."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.prune(max_sigma=2)
+    assert len(atom.transform(X_bin)) == len(X_bin)
+
+
+def test_transform_verbose_invalid():
+    """Assert an error is raised for an invalid value of verbose."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.clean()
+    pytest.raises(ValueError, atom.transform, X_bin, verbose=3)
+
+
+def test_transform_with_y():
+    """Assert that the transform method works when y is provided."""
+    atom = ATOMClassifier(X10_nan, y10, random_state=1)
+    atom.impute(strat_num="drop")
+    X, y = atom.transform(X10_nan, y10)
+    assert len(y) < len(y10)
+
+
+def test_transform_skip_y():
+    """Assert that only transformers are skipped when not provided."""
+    atom = ATOMClassifier(X10_str, y10_str, random_state=1)
+    atom.add(LabelEncoder())
+    atom.encode()
+    assert isinstance(atom.transform(X10_str), pd.DataFrame)
+
+
+# Test base transformers =========================================== >>
+
 def test_custom_params_to_method():
     """Assert that a custom parameter is passed to the method."""
     atom = ATOMClassifier(X_bin, y_bin, verbose=1, random_state=1)
@@ -465,18 +479,137 @@ def test_custom_params_to_method():
     assert atom.pipeline[0].verbose == 2
 
 
-def test_drop_target_column():
-    """Assert that an error is raised when the target column is dropped."""
+def test_add_depending_models():
+    """Assert that an error is raised when the branch has dependent models."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    pytest.raises(ValueError, atom.drop, columns=-1)
+    atom.run("LR")
+    with pytest.raises(PermissionError, match=r".*allowed to add transformers.*"):
+        atom.clean()
 
 
-def test_drop():
-    """Assert that columns can be dropped through the pipeline."""
+def test_add_no_transformer():
+    """Assert that an error is raised if the estimator has no estimator."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.drop([0, 1])
-    assert atom.n_features == X_bin.shape[1] - 2
-    assert str(atom.pipeline[0]).startswith("DropTransformer(columns")
+    with pytest.raises(AttributeError, match=r".*should have a transform method.*"):
+        atom.add(RandomForestClassifier())
+
+
+def test_add_basetransformer_params_are_attached():
+    """Assert that the n_jobs and random_state params from atom are used."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.add(PCA())  # When left to default
+    atom.add(PCA(random_state=2))  # When changed
+    assert atom.pipeline[0].get_params()["random_state"] == 1
+    assert atom.pipeline[1].get_params()["random_state"] == 2
+
+
+def test_add_train_only():
+    """Assert that atom accepts transformers for the train set only."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.add(StandardScaler(), train_only=True)
+    assert check_scaling(atom.X_train) and not check_scaling(atom.X_test)
+
+    len_train, len_test = len(atom.train), len(atom.test)
+    atom.add(Pruner(), train_only=True)
+    assert len(atom.train) != len_train and len(atom.test) == len_test
+
+
+def test_add_complete_dataset():
+    """Assert that atom accepts transformers for the complete dataset."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.add(StandardScaler())
+    assert check_scaling(atom.dataset)
+
+    len_dataset = len(atom.dataset)
+    atom.add(Pruner())
+    assert len(atom.dataset) != len_dataset
+
+
+def test_add_transformer_only_y():
+    """Assert that atom accepts transformers with only y."""
+    atom = ATOMClassifier(X10, y10_str, random_state=1)
+    atom.add(LabelEncoder())
+    assert np.all((atom["target"] == 0) | (atom["target"] == 1))
+
+
+def test_returned_column_already_exists():
+    """Assert that an error is raised if an existing column is returned."""
+    atom = ATOMClassifier(X_text, y10, random_state=1)
+    atom.apply(lambda x: 1, columns="new")
+    with pytest.raises(RuntimeError, match=r".*already exists in the original.*"):
+        atom.vectorize(columns="corpus")
+
+
+def test_add_sparse_matrices():
+    """Assert that transformers that return sp.matrix are accepted."""
+    atom = ATOMClassifier(X10_str, y10, random_state=1)
+    atom.add(OneHotEncoder(handle_unknown="ignore"), columns=2)
+    assert atom.shape == (10, 8)  # Creates 4 extra columns
+
+
+def test_add_keep_column_names():
+    """Assert that the column names are kept after transforming."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+
+    # When the columns are only transformed
+    cols = atom.columns.copy()
+    atom.add(StandardScaler())
+    assert atom.columns == cols
+
+    # When columns were removed
+    cols = atom.columns.copy()
+    atom.add(SelectFromModel(RandomForestClassifier()))
+    assert all(col in cols for col in atom.columns)
+
+
+def test_raise_length_mismatch():
+    """Assert that an error is raised when there's a mismatch in row length."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    with pytest.raises(ValueError, match=r".*does not match length.*"):
+        atom.prune(columns=[2, 4])
+
+
+def test_add_derivative_columns_keep_position():
+    """Assert that derivative columns go after the original."""
+    atom = ATOMClassifier(X10_str, y10, random_state=1)
+    atom.encode(columns="feature 3")
+    assert atom.columns[2:5] == ["feature 3_a", "feature 3_b", "feature 3_c"]
+
+
+def test_add_sets_are_kept_equal():
+    """Assert that the train and test sets always keep the same rows."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    len_train, len_test = len(atom.train), len(atom.test)
+    atom.add(Pruner())
+    assert len(atom.train) < len_train and len(atom.test) < len_test
+
+
+def test_add_reset_index():
+    """Assert that the indices are reset when index=False."""
+    atom = ATOMClassifier(X_bin, y_bin, index=False, random_state=1)
+    atom.prune()
+    assert list(atom.dataset.index) == list(range(len(atom.dataset)))
+
+
+def test_add_params_to_method():
+    """Assert that atom's parameters are passed to the method."""
+    atom = ATOMClassifier(X_bin, y_bin, verbose=1, random_state=1)
+    atom.scale()
+    assert atom.pipeline[0].verbose == 1
+
+
+def test_add_pipeline():
+    """Assert that adding a pipeline adds every individual step."""
+    pipeline = Pipeline(
+        steps=[
+            ("scaler", StandardScaler()),
+            ("sfm", SelectFromModel(RandomForestClassifier())),
+        ]
+    )
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.add(pipeline)
+    assert isinstance(atom.pipeline[0], StandardScaler)
+    assert isinstance(atom.pipeline[1], SelectFromModel)
 
 
 def test_apply_not_callable():
@@ -511,119 +644,12 @@ def test_apply_args_and_kwargs():
     assert atom["new column"][0] == atom["mean texture"][0] + 10
 
 
-def test_add_pipeline():
-    """Assert that adding a pipeline adds every individual step."""
-    pipeline = Pipeline(
-        steps=[
-            ("scaler", StandardScaler()),
-            ("sfm", SelectFromModel(RandomForestClassifier())),
-        ]
-    )
+def test_drop():
+    """Assert that columns can be dropped through the pipeline."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.add(pipeline)
-    assert isinstance(atom.pipeline[0], StandardScaler)
-    assert isinstance(atom.pipeline[1], SelectFromModel)
-
-
-def test_no_transformer():
-    """Assert that an error is raised if the estimator has no transformer."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    pytest.raises(ValueError, atom.add, RandomForestClassifier())
-
-
-def test_basetransformer_params_are_attached():
-    """Assert that the n_jobs and random_state params from atom are used."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.add(PCA())  # When left to default
-    atom.add(PCA(random_state=2))  # When changed
-    assert atom.pipeline[0].get_params()["random_state"] == 1
-    assert atom.pipeline[1].get_params()["random_state"] == 2
-
-
-def test_add_train_only():
-    """Assert that atom accepts transformers for the train set only."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.add(StandardScaler(), train_only=True)
-    assert check_scaling(atom.X_train) and not check_scaling(atom.X_test)
-
-    len_train, len_test = len(atom.train), len(atom.test)
-    atom.add(Pruner(), train_only=True)
-    assert len(atom.train) != len_train and len(atom.test) == len_test
-
-
-def test_add_complete_dataset():
-    """Assert that atom accepts transformers for the complete dataset."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.add(StandardScaler())
-    assert check_scaling(atom.dataset)
-
-    len_dataset = len(atom.dataset)
-    atom.add(Pruner())
-    assert len(atom.dataset) != len_dataset
-
-
-def test_transformer_only_y():
-    """Assert that atom accepts transformers with only an y parameter."""
-    atom = ATOMClassifier(X10, y10_str, random_state=1)
-    atom.add(LabelEncoder())
-    assert np.all((atom["Target"] == 0) | (atom["Target"] == 1))
-
-
-def test_keep_column_names():
-    """Assert that the column names are kept after transforming."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-
-    # When the columns are only transformed
-    cols = atom.columns.copy()
-    atom.add(StandardScaler())
-    assert atom.columns == cols
-
-    # When columns were removed
-    cols = atom.columns.copy()
-    atom.add(SelectFromModel(RandomForestClassifier()))
-    assert all(col in cols for col in atom.columns)
-
-
-def test_subset_columns():
-    """Assert that you can use a subset of the columns."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-
-    # Column indices
-    cols = atom.columns.copy()
-    atom.scale(columns=[3, 4])
-    assert atom.columns == cols  # All columns are kept
-    assert check_scaling(atom.X.iloc[:, [3, 4]])
-    assert not check_scaling(atom.dataset.iloc[:, [7, 8]])
-
-    # Column names
-    atom.scale(columns=["mean radius", "mean texture"])
-    assert check_scaling(atom.dataset.iloc[:, [0, 1]])
-
-    # Column slice
-    atom.scale(columns=slice(10, 12))
-    assert check_scaling(atom.dataset.iloc[:, [10, 11]])
-
-
-def test_derivative_columns_keep_position():
-    """Assert that derivative columns go after the original."""
-    atom = ATOMClassifier(X10_str, y10, random_state=1)
-    atom.encode()
-    assert atom.columns[2].startswith("Feature 3")
-
-
-def test_duplicate_columns_are_ignored():
-    """Assert that duplicate columns are ignored for the transformers."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.add(StandardScaler(), columns=["mean radius", "mean radius"])
-    assert not atom["mean radius"].equals(X_bin["mean radius"])
-
-
-def test_sets_are_kept_equal():
-    """Assert that the train and test sets always keep the same rows."""
-    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    len_train, len_test = len(atom.train), len(atom.test)
-    atom.add(Pruner())
-    assert len(atom.train) < len_train and len(atom.test) < len_test
+    atom.drop([0, 1])
+    assert atom.n_features == X_bin.shape[1] - 2
+    assert str(atom.pipeline[0]).startswith("DropTransformer(columns")
 
 
 # Test data cleaning transformers =================================== >>
@@ -647,10 +673,10 @@ def test_gauss():
 
 def test_clean():
     """Assert that the clean method cleans the dataset."""
-    atom = ATOMClassifier(X10, y10_sn, random_state=1)
+    atom = ATOMClassifier(X10, y10_sn, stratify=False, random_state=1)
     atom.clean()
     assert len(atom.dataset) == 9
-    assert atom.mapping == {"n": 0, "y": 1}
+    assert atom.mapping == {"target": {"n": 0, "y": 1}}
 
 
 def test_impute():
@@ -660,11 +686,18 @@ def test_impute():
     assert atom.dataset.isna().sum().sum() == 0
 
 
+def test_discretize():
+    """Assert that the discretize method bins the numerical columns."""
+    atom = ATOMClassifier(X_bin, y_bin, random_state=1)
+    atom.discretize()
+    assert all(dtype.name == "object" for dtype in atom.X.dtypes)
+
+
 def test_encode():
     """Assert that the encode method encodes all categorical columns."""
     atom = ATOMClassifier(X10_str, y10, random_state=1)
     atom.encode()
-    assert all([atom.X[col].dtype.kind in "ifu" for col in atom.X.columns])
+    assert all(atom.X[col].dtype.kind in "ifu" for col in atom.X.columns)
 
 
 def test_prune():
@@ -688,7 +721,6 @@ def test_balance():
     length = (atom.y_train == 1).sum()
     atom.balance(strategy="NearMiss")
     assert (atom.y_train == 1).sum() != length
-    assert atom.pipeline[0].mapping == atom.mapping
     assert hasattr(atom, "nearmiss")
 
 
@@ -696,32 +728,32 @@ def test_balance():
 
 def test_textclean():
     """Assert that the textclean method cleans the corpus."""
-    atom = ATOMClassifier(X_text, y_text, random_state=1)
+    atom = ATOMClassifier(X_text, y10, shuffle=False, random_state=1)
     atom.textclean()
-    assert atom["Corpus"][0] == "yes sir "
+    assert atom["corpus"][0] == "i am in new york"
     assert hasattr(atom, "drops")
 
 
 def test_tokenize():
     """Assert that the tokenize method tokenizes the corpus."""
-    atom = ATOMClassifier(X_text, y_text, random_state=1)
+    atom = ATOMClassifier(X_text, y10, shuffle=False, random_state=1)
     atom.tokenize()
-    assert atom["Corpus"][0] == ["yes", "sir", "12"]
+    assert atom["corpus"][0] == ["I", "àm", "in", "ne", "'", "w", "york"]
 
 
 def test_normalize():
     """Assert that the normalize method normalizes the corpus."""
-    atom = ATOMClassifier(X_text, y_text, random_state=1)
+    atom = ATOMClassifier(X_text, y10, shuffle=False, random_state=1)
     atom.normalize(stopwords=False, custom_stopwords=["yes"])
-    assert atom["Corpus"][0] == ["sir", "12"]
+    assert atom["corpus"][0] == ["I", "àm", "in", "ne'w", "york"]
 
 
 def test_vectorize():
-    """Assert that the vectorize method vectorizes the corpus."""
-    atom = ATOMClassifier(X_text, y_text, random_state=1)
+    """Assert that the vectorize method converts the corpus to numerical."""
+    atom = ATOMClassifier(X_text, y10, test_size=0.25, random_state=1)
     atom.vectorize(strategy="hashing", n_features=5)
-    assert "Corpus" not in atom
-    assert atom.shape == (4, 6)
+    assert "corpus" not in atom
+    assert atom.shape == (10, 6)
     assert hasattr(atom, "hashing")
 
 
@@ -729,22 +761,22 @@ def test_vectorize():
 
 def test_feature_extraction():
     """Assert that the feature_extraction method creates datetime features."""
-    atom = ATOMClassifier(X10_dt, y10, random_state=1)
-    atom.feature_extraction()
+    atom = ATOMClassifier(X10_dt, y10, verbose=2, random_state=1)
+    atom.feature_extraction(fmt="%d/%m/%Y")
     assert atom.X.shape[1] == 6
 
 
 def test_feature_generation():
     """Assert that the feature_generation method creates extra features."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.feature_generation(n_features=2, generations=5, population=200)
+    atom.feature_generation(n_features=2)
     assert atom.X.shape[1] == X_bin.shape[1] + 2
 
 
 def test_feature_generation_attributes():
     """Assert that the attrs from feature_generation are passed to atom."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.feature_generation("GFG", n_features=2, generations=5, population=200)
+    atom.feature_generation("GFG", n_features=2)
     assert hasattr(atom, "symbolic_transformer")
     assert hasattr(atom, "genetic_features")
 
@@ -762,20 +794,21 @@ def test_default_solver_univariate():
     # For classification tasks
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.feature_selection(strategy="univariate", solver=None, n_features=8)
-    assert atom.pipeline[0].solver.__name__ == "f_classif"
+    assert atom.pipeline[0]._solver.__name__ == "f_classif"
 
     # For regression tasks
     atom = ATOMRegressor(X_reg, y_reg, random_state=1)
     atom.feature_selection(strategy="univariate", solver=None, n_features=8)
-    assert atom.pipeline[0].solver.__name__ == "f_regression"
+    assert atom.pipeline[0]._solver.__name__ == "f_regression"
 
 
 def test_winner_solver_after_run():
     """Assert that the solver is the winning model after run."""
     atom = ATOMClassifier(X_class, y_class, random_state=1)
     atom.run("LR")
+    atom.branch = "fs_branch"
     atom.feature_selection(strategy="SFM", solver=None, n_features=8)
-    assert atom.pipeline[0].solver is atom.winner.estimator
+    assert atom.pipeline[0]._solver is atom.winner.estimator
 
 
 def test_default_solver_from_task():
@@ -783,12 +816,12 @@ def test_default_solver_from_task():
     # For classification tasks
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.feature_selection(strategy="rfe", solver="lgb", n_features=8)
-    assert type(atom.pipeline[0].solver).__name__ == "LGBMClassifier"
+    assert atom.pipeline[0]._solver.__class__.__name__ == "LGBMClassifier"
 
     # For regression tasks
     atom = ATOMRegressor(X_reg, y_reg, random_state=1)
     atom.feature_selection(strategy="rfe", solver="lgb", n_features=25)
-    assert type(atom.pipeline[0].solver).__name__ == "LGBMRegressor"
+    assert atom.pipeline[0]._solver.__class__.__name__ == "LGBMRegressor"
 
 
 @patch("atom.feature_engineering.SequentialFeatureSelector")
@@ -796,6 +829,7 @@ def test_default_scoring(cls):
     """Assert that the scoring is atom's metric when exists."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
     atom.run("lr", metric="recall")
+    atom.branch = "fs_branch"
     atom.feature_selection(strategy="sfs", solver="lgb", n_features=25)
     assert atom.pipeline[0].kwargs["scoring"].name == "recall"
 
@@ -849,7 +883,7 @@ def test_models_and_metric_are_updated():
 def test_errors_are_removed():
     """Assert that the errors are removed if subsequent runs are successful."""
     atom = ATOMClassifier(X_bin, y_bin, random_state=1)
-    atom.run(["BNB", "Tree"], bo_params={"dimensions": {"Tree": 2}})  # Invalid dims
+    atom.run(["BNB", "Tree"], n_initial_points=(2, 7))  # Invalid value for Tree
     atom.run("Tree")  # Runs correctly
     assert not atom.errors  # Errors should be empty
 
