@@ -11,6 +11,10 @@ Description: Unit tests for data_cleaning.py
 import pytest
 import numpy as np
 import pandas as pd
+from imblearn.combine import SMOTETomek
+from sklearn.preprocessing import StandardScaler
+from sklearn.ensemble import RandomForestClassifier
+from category_encoders.leave_one_out import LeaveOneOutEncoder
 
 # Own modules
 from atom.data_cleaning import (
@@ -18,6 +22,7 @@ from atom.data_cleaning import (
     Gauss,
     Cleaner,
     Imputer,
+    Discretizer,
     Encoder,
     Pruner,
     Balancer,
@@ -27,8 +32,8 @@ from atom.utils import (
     NotFittedError, check_scaling,
 )
 from .utils import (
-    X_bin, y_bin, X_class, y_class, X10, X10_nan, X10_str,
-    X10_str2, X10_sn, y10, y10_str,
+    X_bin, y_bin, X_class, y_class, X_idx, y_idx, X10, X10_nan,
+    X10_str, X10_str2, X10_sn, y10, y10_str,
 )
 
 
@@ -60,7 +65,8 @@ def test_scaler_check_is_fitted():
 def test_scaler_invalid_strategy():
     """Assert that an error is raised when strategy is invalid."""
     scaler = Scaler(strategy="invalid")
-    pytest.raises(ValueError, scaler.fit, X_bin)
+    with pytest.raises(ValueError, match=r".*value for the strategy.*"):
+        scaler.fit(X_bin)
 
 
 @pytest.mark.parametrize("strategy", SCALING_STRATS)
@@ -114,7 +120,8 @@ def test_gauss_check_is_fitted():
 def test_invalid_strategy():
     """Assert that an error is raised when strategy is invalid."""
     gauss = Gauss(strategy="invalid")
-    pytest.raises(ValueError, gauss.fit, X_bin)
+    with pytest.raises(ValueError, match=r".*value for the strategy.*"):
+        gauss.fit(X_bin)
 
 
 @pytest.mark.parametrize("strategy", ["yeo-johnson", "box-cox", "quantile"])
@@ -154,13 +161,6 @@ def test_gauss_attach_attribute():
 
 # Test Cleaner ==================================================== >>
 
-def test_invalid_type_ignored():
-    """Assert that prohibited types is ignored when None."""
-    cleaner = Cleaner(drop_types=None)
-    cleaner.transform(X_bin)
-    assert cleaner.drop_types == []
-
-
 def test_drop_invalid_column_type():
     """Assert that invalid columns types are dropped for string input."""
     X = X_bin.copy()
@@ -252,19 +252,22 @@ def test_target_mapping():
 def test_strat_num_parameter():
     """Assert that the strat_num parameter is set correctly."""
     imputer = Imputer(strat_num="invalid")
-    pytest.raises(ValueError, imputer.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*Unknown strategy for the strat_num.*"):
+        imputer.fit(X_bin, y_bin)
 
 
 def test_invalid_max_nan_rows():
     """Assert that an error is raised for invalid max_nan_rows."""
     imputer = Imputer(max_nan_rows=-2)
-    pytest.raises(ValueError, imputer.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the max_nan_rows.*"):
+        imputer.fit(X_bin, y_bin)
 
 
 def test_invalid_max_nan_cols():
     """Assert that an error is raised for invalid max_nan_cols."""
     imputer = Imputer(max_nan_cols=-5)
-    pytest.raises(ValueError, imputer.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the max_nan_cols.*"):
+        imputer.fit(X_bin, y_bin)
 
 
 def test_imputer_check_is_fitted():
@@ -293,25 +296,35 @@ def test_imputing_all_missing_values_categorical(missing):
     assert X.isna().sum().sum() == 0
 
 
-def test_rows_too_many_nans():
+@pytest.mark.parametrize("max_nan_rows", [5, 0.5])
+def test_rows_too_many_nans(max_nan_rows):
     """Assert that rows with too many missing values are dropped."""
     X = X_bin.copy()
     for i in range(5):  # Add 5 rows with all NaN values
         X.loc[len(X)] = [np.nan for _ in range(X.shape[1])]
     y = [np.random.randint(2) for _ in range(len(X))]
-    impute = Imputer(strat_num="mean", strat_cat="most_frequent", max_nan_rows=0.5)
-    X, y = impute.fit_transform(X, y)
+    imputer = Imputer(
+        strat_num="mean",
+        strat_cat="most_frequent",
+        max_nan_rows=max_nan_rows,
+    )
+    X, y = imputer.fit_transform(X, y)
     assert len(X) == 569  # Original size
     assert X.isna().sum().sum() == 0
 
 
-def test_cols_too_many_nans():
+@pytest.mark.parametrize("max_nan_cols", [20, 0.5])
+def test_cols_too_many_nans(max_nan_cols):
     """Assert that columns with too many missing values are dropped."""
     X = X_bin.copy()
     for i in range(5):  # Add 5 cols with all NaN values
         X["col " + str(i)] = [np.nan for _ in range(X.shape[0])]
-    impute = Imputer(strat_num="mean", strat_cat="most_frequent", max_nan_cols=0.5)
-    X, y = impute.fit_transform(X, y_bin)
+    imputer = Imputer(
+        strat_num="mean",
+        strat_cat="most_frequent",
+        max_nan_cols=max_nan_cols,
+    )
+    X, y = imputer.fit_transform(X, y_bin)
     assert len(X.columns) == 30  # Original number of columns
     assert X.isna().sum().sum() == 0
 
@@ -388,38 +401,151 @@ def test_imputing_non_numeric_most_frequent():
     assert X.isna().sum().sum() == 0
 
 
+# Test Discretizer ================================================= >>
+
+def test_strategy_parameter_discretizer():
+    """Assert that the strategy parameter is set correctly."""
+    discretizer = Discretizer(strategy="invalid")
+    with pytest.raises(ValueError, match=r".*value for the strategy parameter.*"):
+        discretizer.fit(X_bin)
+
+
+def test_invalid_bins_missing_column():
+    """Assert that an error is raised when a column is missing."""
+    discretizer = Discretizer(strategy="uniform", bins={"invalid": 5})
+    with pytest.raises(ValueError, match=r".*not found in the dictionary.*"):
+        discretizer.fit(X_bin)
+
+
+def test_invalid_bins_custom_strategy():
+    """Assert that an error is raised when bins are not a sequence."""
+    discretizer = Discretizer(strategy="custom", bins=5)
+    with pytest.raises(TypeError, match=r".*a sequence of bin edges.*"):
+        discretizer.fit(X_bin)
+
+
+def test_invalid_length_labels():
+    """Assert that an error is raised when len(bins) != len(labels)."""
+    discretizer = Discretizer(strategy="custom", bins=[5, 10, 15], labels=["label"])
+    with pytest.raises(ValueError, match=r".*length of the bins does not match.*"):
+        discretizer.fit(X_bin)
+
+
+def test_invalid_bins_to_column_length():
+    """Assert that an error is raised when len(bins) != len(columns)."""
+    discretizer = Discretizer(strategy="uniform", bins=[5, 10])
+    with pytest.raises(ValueError, match=r".*length of the bins does not match.*"):
+        discretizer.fit(X_bin)
+
+
+@pytest.mark.parametrize("strategy", ["uniform", "quantile", "kmeans"])
+def test_discretizer_strategies(strategy):
+    """Assert that custom binning can be performed."""
+    discretizer = Discretizer(strategy=strategy, bins=5)
+    X = discretizer.fit_transform(X_bin)
+    assert all(X[col].dtype.name == "object" for col in X)
+
+
+def test_custom_strategy():
+    """Assert that custom binning can be performed."""
+    discretizer = Discretizer(strategy="custom", bins=[0, 25, 50])
+    X = discretizer.fit_transform(X_bin)
+    assert X["mean texture"].unique().tolist() == ["0-25", "25-50"]
+
+
+def test_bins_is_sequence():
+    """Assert that bins can be provided as sequence."""
+    discretizer = Discretizer(strategy="uniform", bins=[5, 6, 7])
+    X = discretizer.fit_transform(X_bin.iloc[:, :3])
+    assert X[X.columns[0]].nunique() == 5
+    assert X[X.columns[1]].nunique() == 6
+    assert X[X.columns[2]].nunique() == 7
+
+
+def test_bins_is_dict():
+    """Assert that bins can be provided as dict."""
+    discretizer = Discretizer(
+        strategy="uniform",
+        bins={X_bin.columns[0]: 5, X_bin.columns[1]: 6, X_bin.columns[2]: 7},
+    )
+    X = discretizer.fit_transform(X_bin.iloc[:, :3])
+    assert X[X.columns[0]].nunique() == 5
+    assert X[X.columns[1]].nunique() == 6
+    assert X[X.columns[2]].nunique() == 7
+
+
+def test_labels_non_custom_strategy():
+    """Assert that custom labels can be added to strategy != custom."""
+    discretizer = Discretizer(strategy="uniform", bins=3, labels=["l1", "l2", "l3"])
+    X = discretizer.fit_transform(X_bin)
+    assert X["mean texture"].unique().tolist() == ["l1", "l2", "l3"]
+
+
+def test_labels_custom_strategy():
+    """Assert that custom labels can be added to the custom strategy."""
+    discretizer = Discretizer(
+        strategy="custom",
+        bins=[0, 10, 20, 50],
+        labels={"mean texture": ["l1", "l2", "l3"]},
+    )
+    X = discretizer.fit_transform(X_bin)
+    assert X["mean texture"].unique().tolist() == ["l2", "l3", "l1"]
+
+
 # Test Encoder ===================================================== >>
 
 def test_strategy_parameter_encoder():
     """Assert that the strategy parameter is set correctly."""
     encoder = Encoder(strategy="invalid")
-    pytest.raises(ValueError, encoder.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the strategy.*"):
+        encoder.fit(X10_str, y10)
 
 
 def test_strategy_with_encoder_at_end():
     """Assert that the strategy works with Encoder at the end of the string."""
     encoder = Encoder(strategy="TargetEncoder", max_onehot=None)
     encoder.fit(X10_str, y10)
-    assert encoder._encoders["Feature 3"].__class__.__name__ == "TargetEncoder"
+    assert encoder._encoders["feature 3"].__class__.__name__ == "TargetEncoder"
 
 
 def test_max_onehot_parameter():
     """Assert that the max_onehot parameter is set correctly."""
     encoder = Encoder(max_onehot=-2)
-    pytest.raises(ValueError, encoder.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the max_onehot.*"):
+        encoder.fit(X10_str, y10)
 
 
 def test_frac_to_other_parameter():
     """Assert that the frac_to_other parameter is set correctly."""
     encoder = Encoder(frac_to_other=-2)
-    pytest.raises(ValueError, encoder.fit, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the frac_to_other.*"):
+        encoder.fit(X10_str, y10)
 
 
-def test_frac_to_other():
+@pytest.mark.parametrize("frac_to_other", [3, 0.3])
+def test_frac_to_other(frac_to_other):
     """Assert that the other values are created when encoding."""
-    encoder = Encoder(max_onehot=5, frac_to_other=0.3)
+    encoder = Encoder(max_onehot=5, frac_to_other=frac_to_other)
     X = encoder.fit_transform(X10_str, y10)
-    assert "Feature 3_other" in X.columns
+    assert "feature 3_other" in X.columns
+
+
+def test_encoder_strategy_invalid_estimator():
+    """Assert that an error is raised when strategy is invalid."""
+    encoder = Encoder(strategy=RandomForestClassifier())
+    with pytest.raises(TypeError, match=r".*type for the strategy.*"):
+        encoder.fit_transform(X10_str, y10)
+
+
+def test_encoder_custom_estimator():
+    """Assert that the strategy can be a custom estimator."""
+    encoder = Encoder(strategy=LeaveOneOutEncoder, max_onehot=None)
+    X = encoder.fit_transform(X10_str, y10)
+    assert X.loc[0, "feature 3"] != "a"
+
+    encoder = Encoder(strategy=LeaveOneOutEncoder(), max_onehot=None)
+    X = encoder.fit_transform(X10_str, y10)
+    assert X.loc[0, "feature 3"] != "a"
 
 
 def test_encoder_check_is_fitted():
@@ -428,28 +554,29 @@ def test_encoder_check_is_fitted():
 
 
 def test_missing_values_are_propagated():
-    """Assert that an missing are propagated."""
+    """Assert that missing values are propagated."""
     encoder = Encoder(max_onehot=None)
     assert np.isnan(encoder.fit_transform(X10_sn, y10).iloc[0, 2])
 
 
-def test_unknown_classes():
-    """Assert that unknown classes are converted to NaN."""
+def test_unknown_classes_are_imputed():
+    """Assert that unknown classes are imputed."""
     encoder = Encoder()
     encoder.fit(["a", "b", "b", "a"])
-    assert encoder.transform(["c"]).isna().sum().sum() == 1
+    assert encoder.transform(["c"]).iloc[0, 0] == -1.0
 
 
 def test_ordinal_encoder():
     """Assert that the Ordinal-encoder works as intended."""
     encoder = Encoder(max_onehot=None)
     X = encoder.fit_transform(X10_str2, y10)
-    assert np.all((X["Feature 3"] == 0) | (X["Feature 3"] == 1))
+    assert np.all((X["feature 3"] == 0) | (X["feature 3"] == 1))
+    assert list(encoder.mapping) == ["feature 3", "feature 4"]
 
 
 def test_ordinal_features():
     """Assert that ordinal features are encoded."""
-    encoder = Encoder(max_onehot=None, ordinal={"Feature 3": ["b", "a"]})
+    encoder = Encoder(max_onehot=None, ordinal={"feature 3": ["b", "a"]})
     X = encoder.fit_transform(X10_str2, y10)
     assert X.iloc[0, 2] == 1 and X.iloc[2, 2] == 0
 
@@ -458,7 +585,7 @@ def test_one_hot_encoder():
     """Assert that the OneHot-encoder works as intended."""
     encoder = Encoder(max_onehot=4)
     X = encoder.fit_transform(X10_str, y10)
-    assert "Feature 3_c" in X.columns
+    assert "feature 3_c" in X.columns
 
 
 @pytest.mark.parametrize("strategy", ENCODING_STRATS)
@@ -466,14 +593,14 @@ def test_all_encoder_types(strategy):
     """Assert that all estimators work as intended."""
     encoder = Encoder(strategy=strategy, max_onehot=None)
     X = encoder.fit_transform(X10_str, y10)
-    assert all([X[col].dtype.kind in "ifu" for col in X])
+    assert all(X[col].dtype.kind in "ifu" for col in X)
 
 
 def test_kwargs_parameters():
     """Assert that the kwargs parameter works as intended."""
     encoder = Encoder(strategy="LeaveOneOut", max_onehot=None, sigma=0.5)
     encoder.fit(X10_str, y10)
-    assert encoder._encoders["Feature 3"].get_params()["sigma"] == 0.5
+    assert encoder._encoders["feature 3"].get_params()["sigma"] == 0.5
 
 
 # Test Pruner ====================================================== >>
@@ -481,25 +608,29 @@ def test_kwargs_parameters():
 def test_invalid_strategy_parameter():
     """Assert that an error is raised for an invalid strategy parameter."""
     pruner = Pruner(strategy="invalid")
-    pytest.raises(ValueError, pruner.transform, X_bin)
+    with pytest.raises(ValueError, match=r".*value for the strategy.*"):
+        pruner.transform(X_bin)
 
 
 def test_invalid_method_for_non_z_score():
     """Assert that an error is raised for an invalid method and strat combination."""
     pruner = Pruner(strategy="iforest", method="min_max")
-    pytest.raises(ValueError, pruner.transform, X_bin)
+    with pytest.raises(ValueError, match=r".*accepts another method.*"):
+        pruner.transform(X_bin)
 
 
 def test_invalid_method_parameter():
     """Assert that an error is raised for an invalid method parameter."""
     pruner = Pruner(method="invalid")
-    pytest.raises(ValueError, pruner.transform, X_bin)
+    with pytest.raises(ValueError, match=r".*value for the method parameter.*"):
+        pruner.transform(X_bin)
 
 
 def test_invalid_max_sigma_parameter():
     """Assert that an error is raised for an invalid max_sigma parameter."""
     pruner = Pruner(max_sigma=0)
-    pytest.raises(ValueError, pruner.transform, X_bin)
+    with pytest.raises(ValueError, match=r".*value for the max_sigma.*"):
+        pruner.transform(X_bin)
 
 
 def test_max_sigma_functionality():
@@ -542,7 +673,7 @@ def test_categorical_cols_are_ignored():
     """Assert that categorical columns are returned untouched."""
     Feature_2 = np.array(X10_str)[:, 2]
     X, y = Pruner(method="min_max", max_sigma=2).transform(X10_str, y10)
-    assert [i == j for i, j in zip(X["Feature 2"], Feature_2)]
+    assert [i == j for i, j in zip(X["feature 2"], Feature_2)]
 
 
 def test_drop_outlier_in_target():
@@ -552,7 +683,7 @@ def test_drop_outlier_in_target():
 
 
 @pytest.mark.parametrize("strategy", PRUNING_STRATS)
-def test_strategies(strategy):
+def test_pruner_strategies(strategy):
     """Assert that all estimator requiring strategies work."""
     pruner = Pruner(strategy=strategy)
     X, y = pruner.transform(X_bin, y_bin)
@@ -592,19 +723,39 @@ def test_pruner_attach_attribute():
 
 # Test Balancer ==================================================== >>
 
-def test_strategy_parameter_balancer():
-    """Assert that an error is raised when strategy is invalid."""
+def test_balancer_strategy_unknown_str():
+    """Assert that an error is raised when strategy is unknown."""
     balancer = Balancer(strategy="invalid")
-    pytest.raises(ValueError, balancer.transform, X_bin, y_bin)
+    with pytest.raises(ValueError, match=r".*value for the strategy.*"):
+        balancer.transform(X_bin, y_bin)
 
 
-@pytest.mark.parametrize("strategy", [i for i in BALANCING_STRATS if i != "smotenc"])
-def test_all_balancers(strategy):
-    """Assert that all estimators work as intended."""
-    balancer = Balancer(strategy=strategy, sampling_strategy="all")
+def test_balancer_strategy_invalid_estimator():
+    """Assert that an error is raised when strategy is invalid."""
+    balancer = Balancer(strategy=StandardScaler())
+    with pytest.raises(TypeError, match=r".*type for the strategy.*"):
+        balancer.transform(X_bin, y_bin)
+
+
+def test_balancer_custom_estimator():
+    """Assert that the strategy can be a custom estimator."""
+    balancer = Balancer(strategy=SMOTETomek)
     X, y = balancer.transform(X_bin, y_bin)
     assert len(X) != len(X_bin)
 
+    balancer = Balancer(strategy=SMOTETomek())
+    X, y = balancer.transform(X_bin, y_bin)
+    assert len(X) != len(X_bin)
+
+
+@pytest.mark.parametrize("strategy", BALANCING_STRATS)
+def test_all_balancers(strategy):
+    """Assert that all estimators work as intended."""
+    if strategy not in ("smotenc", "kmeanssmote"):  # Non ATOM related errors
+        balancer = Balancer(strategy=strategy, sampling_strategy="all")
+        X, y = balancer.transform(X_bin, y_bin)
+        assert len(X) != len(X_bin)
+
 
 def test_balancer_kwargs():
     """Assert that kwargs can be passed to the estimator."""
@@ -613,18 +764,36 @@ def test_balancer_kwargs():
     assert balancer.smote.get_params()["k_neighbors"] == 12
 
 
-def test_return_pandas():
-    """Assert that pandas objects are returned, not np.ndarray."""
-    X, y = Balancer().transform(X_bin, y_bin)
-    assert isinstance(X, pd.DataFrame)
-    assert isinstance(y, pd.Series)
+def test_oversampling_numerical_index():
+    """Assert that new samples have an increasing int index."""
+    X, y = Balancer(strategy="smote").transform(X_bin, y_bin)
+    assert list(X.index) == list(range(len(X)))
 
 
-def test_return_correct_column_names():
-    """Assert that the returned objects have the original column names."""
-    X, y = Balancer().transform(X_bin, y_bin)
-    assert list(X.columns) == list(X_bin.columns)
-    assert y.name == y_bin.name
+def test_oversampling_string_index():
+    """Assert that new samples have a new index."""
+    X, y = Balancer(strategy="smote").transform(X_idx, y_idx)
+    assert X.index[-1] == f"smote_{len(X) - len(X_idx)}"
+
+
+def test_undersampling_keeps_indices():
+    """Assert that indices are kept after transformation."""
+    X, y = Balancer(strategy="nearmiss").transform(X_bin, y_bin)
+    assert list(X.index) != list(range(len(X)))
+
+
+def test_combinations_numerical_index():
+    """Assert that new samples have an increasing int index."""
+    X, y = Balancer(strategy="smoteenn").transform(X_bin, y_bin)
+    assert not all(idx in X.index for idx in X_bin.index)  # Samples were dropped
+    assert max(X.index) > max(X_bin.index)  # Samples were added
+
+
+def test_combinations_string_index():
+    """Assert that new samples have a new index."""
+    X, y = Balancer(strategy="smotetomek").transform(X_idx, y_idx)
+    assert not all(idx in X.index for idx in X_bin.index)  # Samples were dropped
+    assert len(X.index.str.startswith("smotetomek") > 0)  # Samples were added
 
 
 def test_balancer_attach_attribute():
